/**
 * This file implements the main routine for the CSOCP solver
 */

#include "csocp.h"

#include "csocp_attr.h"
#include "csocp_cone.h"
#include "csocp_config.h"
#include "csocp_iter.h"
#include "csocp_kkt.h"
#include "csocp_param.h"
#include "csocp_scaling.h"
#include "csocp_solver.h"

#include "tool_buffer.h"
#include "tool_interrupt.h"
#include "tool_math.h"
#include "tool_memory.h"
#include "tool_print.h"
#include "tool_spmat.h"
#include "tool_time.h"

#include "struct_attr.h"
#include "struct_buffer.h"
#include "struct_cone.h"
#include "struct_kkt.h"
#include "struct_param.h"
#include "struct_prob.h"
#include "struct_spmat.h"

#include <stdio.h>
#include <stdlib.h>

#ifdef MATLAB_MEX_FILE
#include "mex.h"
#endif

/**
 * Methods for setting up problem
 */

/* Create problem */
c_int CSOCP_create(csocp_prob **p_prob) {
  csocp_prob *prob = NULL;

  /* Allocate memory for problem */
  CSOCP_INITZ(prob, 1);

  /* Allocate memory for attributes */
  CSOCP_INIT(prob->attr, 1);
  c_attr_reset(prob->attr);

  CSOCP_INIT(prob->best_attr, 1);
  c_attr_reset(prob->best_attr);

  /* Allocate memory for parameters and initialize */
  CSOCP_INIT(prob->param, 1);
  c_param_reset(prob->param);

  /* Pass back pointer */
  *p_prob = prob;

  return 0;
}

/* Compute problem statistics */
static void prob_calstat(c_attr *attr, /* Input attributes */
                         c_int n,      /* Number of variables */
                         c_int m,      /* Number of inequalities */
                         c_int p,      /* Number of equalities */
                         c_int l,      /* Number of LP cone */
                         c_int ncones, /* Number of SOC cones */
                         c_int *q,     /* Dimension of SOC cones */
                         c_int nnzA,   /* Number of nonzeros of A */
                         c_int nnzG    /* Number of nonzeros of G */
) {
  c_int i = 0;
  c_int nK = 0, nnzK = 0;
  c_int nTmpI = 0, nTmpD = 0;

  /* Number of variables and constraints */
  nK = n + p + m;
  /* Number of SOC scalings */
  nK += 2 * ncones;

  /* Number of nonzeros of A */
  nnzK = nnzA;
  /* Number of nonzeros of G */
  nnzK += nnzG;
  /* Number of LP cone */
  nnzK += l;
  /* Number of SOC scalings */
  for (i = 0; i < ncones; i++) {
    nnzK += 3 * q[i] + 1;
  }
  /* Number of static regularization */
  nnzK += n + p;

  /* Maximal length of integer buffer */
  nTmpI = 4 * nK + nnzK + 1;
  /* Maximal length of double buffer */
  nTmpD = CSOCP_MAX(6 * nK, nnzK);

  /* Dimension statistics */
  attr->n = n;
  attr->m = m;
  attr->p = p;
  attr->l = l;
  attr->ncones = ncones;
  for (i = 0; i < ncones; ++i) {
    attr->nconedim += q[i];
  }
  attr->nnzA = nnzA;
  attr->nnzG = nnzG;
  attr->nK = nK;
  attr->nnzK = nnzK;

  attr->nTmpI = nTmpI;
  attr->nTmpD = nTmpD;

#ifdef CSOCP_WITH_STATS
  /* Compute memory usage */
  attr->cone_mem = c_cone_getmem(l, ncones, q);
  attr->total_mem += attr->cone_mem;

  attr->kkt_mem = c_kkt_getmem(n, m, p, nnzA, nnzG, nK, nnzK);
  attr->total_mem += attr->kkt_mem;

  attr->solver_mem = c_solver_getmem(n, m, p, nnzA, nnzG);
  attr->total_mem += attr->solver_mem;

  attr->tmp_mem = c_buffer_getmem(nTmpI, nTmpD);
  attr->total_mem += attr->tmp_mem;

  attr->total_mem += 2 * c_attr_getmem();
  attr->total_mem += c_param_getmem();
#endif
}

#ifdef CSOCP_WITH_STATS
/* Print problem statistics */
static void prob_printstat(c_attr *attr) {
  /* Print dimension summary */
  c_print("\n");

  c_print("Problem summary:\n");
  c_print("  Number of variables:      %d\n", attr->n);
  c_print("  Number of inequalities:   %d\n", attr->m);
  c_print("  Number of equalities:     %d\n", attr->p);
  c_print("  Number of LP cones:       %d\n", attr->l);
  c_print("  Number of SOC cones:      %d\n", attr->ncones);
  c_print("  Sum of SOC dimension:     %d\n", attr->nconedim);
  c_print("  Number of nnzs of A:      %d\n", attr->nnzA);
  c_print("  Number of nnzs of G:      %d\n", attr->nnzG);
  c_print("  Number of cols of KKT:    %d\n", attr->nK);
  c_print("  Number of nnzs of KKT:    %d\n", attr->nnzK);
  c_print("  Number of nnzs of L:      %d\n", attr->nnzL);
  c_print("  Length of integer buffer: %d\n", attr->nTmpI);
  c_print("  Length of double buffer:  %d\n", attr->nTmpD);

  c_print("\n");

  /* Print memory summary */
  c_print("Memory summary:\n");
  c_print("  Size of cone:             %d\n", attr->cone_mem);
  c_print("  Size of KKT:              %d\n", attr->kkt_mem);
  c_print("  Size of solver:           %d\n", attr->solver_mem);
  c_print("  Size of ordering:         %d\n", attr->order_mem);
  c_print("  Size of temporary:        %d\n", attr->tmp_mem);
  c_print("  Size of alloc-free:       %d\n", attr->free_mem);
  c_print("  Size of total:            %d\n", attr->total_mem);
  c_print("\n");

  return;
}
#endif

/* Builds the upper triangular part of KKT matrix */
static c_int prob_setupkkt(csocp_prob *prob) {
  c_int i = 0, j = 0, l = 0, r = 0;
  c_int begidx = 0, endidx = 0;
  c_int colidx = 0, elemidx = 0, coneidx = 0;

  c_spmat *At = NULL;
  c_spmat *Gt = NULL;
  c_spmat *ku = NULL;

  c_int *sign = NULL;
  c_int *kbeg = NULL;
  c_int *kidx = NULL;
  c_real *kelem = NULL;

  c_cone *cone = prob->cone;
  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

  c_attr *attr = prob->attr;
  c_param *param = prob->param;

  c_buffer *buffer = prob->buffer;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting setup KKT matrix\n", __FILE__, __LINE__);
#endif

  /*
   * Setting up KKT matrix, only the upper triangular part
   *
   *   [0   A'  G']
   *   [A   0   0 ]
   *   [G   0  -I ]
   */

  /* Get A' and G' */
  if (prob->A != NULL) {
    At = c_spmat_transpose(prob->A, buffer);
  }
  Gt = c_spmat_transpose(prob->G, buffer);

  /* Get memory for KKT matrix from buffer */
  c_buffer_request_ibuf(buffer, kbeg, attr->nK + 1);
  c_buffer_request_ibuf(buffer, kidx, attr->nnzK);
  c_buffer_request_dbuf(buffer, kelem, attr->nnzK);

  /* (1, 1) block: Regularization values */
  elemidx = 0;
  for (j = 0; j < attr->n; j++) {
    kbeg[j] = j;
    kidx[j] = j;
    kelem[elemidx++] = param->deltastat;
  }

  /* (1, 2) and (2, 2) blocks */
  i = 0;
  for (j = 0; j < attr->p; j++) {
    /* (1, 2) block: A' */
    begidx = At->colBeg[j];
    endidx = At->colBeg[j + 1];
    colidx = attr->n + j;
    if (begidx <= endidx) {
      kbeg[colidx] = elemidx;
      while (begidx++ < endidx) {
        kidx[elemidx] = At->colIdx[i];
        kelem[elemidx] = At->colElem[i];
        i++;
        elemidx++;
      }
    }

    /* (2, 2) block: Regularization values */
    kidx[elemidx] = colidx;
    kelem[elemidx++] = -param->deltastat;
  }

  /* (1, 3) and (3, 3) blocks */
  i = 0;
  for (j = 0; j < lp->ndim; j++) {
    /* (1, 3) block: G' */
    begidx = Gt->colBeg[j];
    endidx = Gt->colBeg[j + 1];
    colidx = attr->n + attr->p + j;
    if (begidx <= endidx) {
      kbeg[colidx] = elemidx;
      while (begidx++ < endidx) {
        kidx[elemidx] = Gt->colIdx[i];
        kelem[elemidx] = Gt->colElem[i];
        i++;
        elemidx++;
      }
    }

    /* Map between LP cone and element index in KKT matrix */
    lp->idxmap[j] = elemidx;

    /* -I for LP cone */
    kidx[elemidx] = colidx;
    kelem[elemidx] = -1.0;
    elemidx++;
  }

  /*
   * Scaling matrix of SOC cones:
   *
   *   [I   v   u]
   *   [u'  1   0]
   *   [v'  0' -1]
   */

  coneidx = lp->ndim;
  for (l = 0; l < cone->nsoc; l++) {
    soc = cone->soc[l];

    /* (3, 3) block: G' */
    for (j = 0; j < soc->ndim; j++) {
      begidx = Gt->colBeg[coneidx + j];
      endidx = Gt->colBeg[coneidx + j + 1];
      colidx = attr->n + attr->p + coneidx + 2 * l + j;
      if (begidx <= endidx) {
        kbeg[colidx] = elemidx;
        while (begidx++ < endidx) {
          kidx[elemidx] = Gt->colIdx[i];
          kelem[elemidx] = Gt->colElem[i];
          i++;
          elemidx++;
        }
      }

      /* Map between diagonal D of SOC cone and element index in KKT matrix */
      soc->idxmap[j] = elemidx;

      /* (3, 3) block: -I for LP */
      kidx[elemidx] = colidx;
      kelem[elemidx] = -1.0;
      elemidx++;
    }

    colidx = attr->n + attr->p + coneidx + 2 * l + soc->ndim;

    /* (3, 3) block: v and -1 */
    kbeg[colidx] = elemidx;
    for (r = 1; r < soc->ndim; r++) {
      kidx[elemidx] = attr->n + attr->p + coneidx + 2 * l + r;
      kelem[elemidx] = 0; // Placeholder
      elemidx++;
    }
    kidx[elemidx] = colidx;
    kelem[elemidx] = -1;
    elemidx++;

    /* (3, 3) block: u and 1 */
    kbeg[colidx + 1] = elemidx;
    for (r = 0; r < soc->ndim; r++) {
      kidx[elemidx] = attr->n + attr->p + coneidx + 2 * l + r;
      kelem[elemidx] = 0; // Placeholder
      elemidx++;
    }
    kidx[elemidx] = colidx + 1;
    kelem[elemidx] = +1;
    elemidx++;

    coneidx += soc->ndim;
  }
  kbeg[attr->nK] = attr->nnzK;

  /* Generate KKT matrix */
  ku = c_spmat_generate(attr->nK, attr->nK, attr->nnzK, kbeg, kidx, kelem);

  /* Free memory */
#ifdef CSOCP_WITH_STATS
  attr->free_mem += c_spmat_getmem(At->nCol, At->nElem);
  attr->free_mem += c_spmat_getmem(Gt->nCol, Gt->nElem);
  attr->total_mem += attr->free_mem;
#endif
  if (At != NULL) {
    c_spmat_free(&At);
  }
  c_spmat_free(&Gt);

  /*
   * Setting up sign vector
   */

  /* Get temporary memory from buffer */
  c_buffer_request_ibuf(buffer, sign, attr->nK);

  /* (1, 1) block */
  c_math_fillint(attr->n, 1, sign);

  /* (2, 2) block */
  c_math_fillint(attr->p, -1, sign + attr->n);

  /* (3, 3) block: LP cone */
  c_math_fillint(lp->ndim, -1, sign + attr->n + attr->p);

  /* (3, 3) block: SOC cone */
  coneidx = attr->n + attr->p + lp->ndim;
  for (l = 0; l < cone->nsoc; l++) {
    soc = cone->soc[l];
    /* D */
    c_math_fillint(soc->ndim, -1, sign + coneidx);
    coneidx += soc->ndim;
    /* v */
    sign[coneidx++] = -1;
    /* u */
    sign[coneidx++] = +1;
  }

  /* Symbolic factorization of KKT system */
  if (c_kkt_symbolic(prob->kkt, ku, sign, buffer, attr) != KKT_OK) {
    return -1;
  }

#ifdef CSOCP_WITH_DUMP
  /* Dump upper triangular KKT matrix to file */
  c_spmat_dump(ku, "ku.txt");
#endif

  /* Release buffer */
  c_buffer_release_ibuf(buffer, attr->nK);     /* sign */
  c_buffer_release_ibuf(buffer, attr->nK + 1); /* kbeg */
  c_buffer_release_ibuf(buffer, attr->nnzK);   /* kidx */
  c_buffer_release_dbuf(buffer, attr->nnzK);   /* kelem */

  /* Free memory */
  CSOCP_FREE(ku);

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish setup KKT matrix (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif

  return 0;
}

/* Set up problem */
c_int CSOCP_loadsocp(csocp_prob *prob, /* The CSOCP problem */
                     c_int n,          /* Number of variables */
                     c_int m,          /* Number of inequalities */
                     c_int p,          /* Number of equalities */
                     c_int l,          /* Number of positive orthant */
                     c_int ncones,     /* Number of SOC cones */
                     c_int *q,         /* Dimension for each SOC cone */
                     c_real *Gpr,      /* Element pointer of inequality matrix */
                     c_int *Gjc,       /* Begin pointer of inequality matrix */
                     c_int *Gir,       /* Index pointer of inequality matrix */
                     c_real *Apr,      /* Element pointer of equality matrix */
                     c_int *Ajc,       /* Begin pointer of equality matrix */
                     c_int *Air,       /* Index pointer of equality matrix */
                     c_real *c,        /* Objective cost */
                     c_real *h,        /* RHS of cones */
                     c_real *b         /* RHS of equalities */
) {
  c_real loadbeg = 0.0;

  /* Start timer for loading process */
  loadbeg = c_timer_elapsed();

#ifdef CSOCP_WITH_DEBUG
  prob->dbgfile = fopen("csocp.log", "a");
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting loading problem\n", __FILE__, __LINE__);
#endif

  /* Assign dimensions of LP/SOC cones */
  prob->n = n;
  prob->m = m;
  prob->p = p;
  prob->degree = l + ncones;

  /* Calculate statistics */
  prob_calstat(prob->attr, n, m, p, l, ncones, q, Ajc ? Ajc[n] : 0, Gjc ? Gjc[n] : 0);

  /* Allocate memory for buffer */
  c_buffer_create(&prob->buffer, prob->attr->nTmpI, prob->attr->nTmpD);

  /* Allocate memory for LP/SOC cones */
#ifdef CSOCP_WITH_DEBUG
  c_cone_create(&prob->cone, l, ncones, q, prob->dbgfile);
#else
  c_cone_create(&prob->cone, l, ncones, q);
#endif

  /* Allocate memory KKT system */
#ifdef CSOCP_WITH_DEBUG
  c_kkt_create(&prob->kkt, prob->attr->nK, prob->attr->nnzK, prob->dbgfile);
#else
  c_kkt_create(&prob->kkt, prob->attr->nK, prob->attr->nnzK);
#endif

  /* Copy problem data */
  if (Apr && Ajc && Air) {
#ifdef CSOCP_WITH_COPYPROB
    prob->A = c_spmat_create(p, n, Ajc[n]);
    c_spmat_loaddata(prob->A, Ajc, Air, Apr);
#else
    prob->A = c_spmat_generate(p, n, Ajc[n], Ajc, Air, Apr);
#endif
  }

  if (Gpr && Gjc && Gir) {
#ifdef CSOCP_WITH_COPYPROB
    prob->G = c_spmat_create(m, n, Gjc[n]);
    c_spmat_loaddata(prob->G, Gjc, Gir, Gpr);
#else
    prob->G = c_spmat_generate(m, n, Gjc[n], Gjc, Gir, Gpr);
#endif
  } else {
    prob->G = c_spmat_generate(0, 0, 0, NULL, NULL, NULL);
  }

  /* Assign costs and RHS of equalities/inequalities */
#ifdef CSOCP_WITH_COPYPROB
  CSOCP_INIT(prob->c, n);
  CSOCP_INIT(prob->h, m);
  CSOCP_INIT(prob->b, p);

  c_math_copydbl(n, c, prob->c, 0);
  c_math_copydbl(m, h, prob->h, 0);
  c_math_copydbl(p, b, prob->b, 0);
#else
  prob->c = c;
  prob->h = h;
  prob->b = b;
#endif

  /* Allocate memory for residuals */
  CSOCP_INIT(prob->rx, n);
  CSOCP_INIT(prob->ry, p);
  CSOCP_INIT(prob->rz, m);

  CSOCP_INIT(prob->dx1, n);
  CSOCP_INIT(prob->dy1, p);
  CSOCP_INIT(prob->dz1, m);

  CSOCP_INIT(prob->dx2, n);
  CSOCP_INIT(prob->dy2, p);
  CSOCP_INIT(prob->dz2, m);

  /* Allocate memory for primal/dual variables */
  CSOCP_INIT(prob->dsaff, m);
  CSOCP_INIT(prob->dzaff, m);
  CSOCP_INIT(prob->dsaff_w, m);
  CSOCP_INIT(prob->dzaff_w, m);

  CSOCP_INIT(prob->lambda, m);

  CSOCP_INIT(prob->x, n);
  CSOCP_INIT(prob->y, p);
  CSOCP_INIT(prob->z, m);
  CSOCP_INIT(prob->s, m);

  /* Allocate memory for storing best solution */
  CSOCP_INIT(prob->best_x, n);
  CSOCP_INIT(prob->best_y, p);
  CSOCP_INIT(prob->best_z, m);
  CSOCP_INIT(prob->best_s, m);

  /* Allocate memory for scaling factors */
  CSOCP_INIT(prob->facX, n);
  CSOCP_INIT(prob->facA, p);
  CSOCP_INIT(prob->facG, m);

  /* Scaling the equality/inequality matrix */
  c_scaling_set(prob);

  /* Setup KKT system */
  if (prob_setupkkt(prob) != 0) {
    prob->attr->status = CSOCP_ERROR;
    return 0;
  }

#ifdef CSOCP_WITH_STATS
  /* Print statistics */
  prob_printstat(prob->attr);
#endif

  /* Get elapsed loading time */
  prob->attr->load_time = c_timer_elapsed() - loadbeg;

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish loading problem (%.4fs)\n", __FILE__,
           __LINE__, prob->attr->load_time);
  c_fprint(prob->dbgfile, "\n");
#endif

  return 0;
}

/**
 * Methods for solving problem
 */

/* Print solving log */
static void solve_printlog(csocp_prob *prob) {
  c_attr *attr = prob->attr;

#ifdef CSOCP_WITH_DEBUG
  if (attr->itercnt == 0) {
    c_print("                  Objective                          Residual\n");
    c_print("Iter       Primal           Dual             Primal            Dual"
            "            Gap           Mu           Step           Sigma\n");
  }
  c_print("%2d    %+10.8e  %+10.8e  %+10.8e  %+10.8e  %+.6e %+.6e %+.6e %+.6e\n",
          (int)attr->itercnt, attr->pobj, attr->dobj, attr->pres, attr->dres,
          attr->absgap, attr->mu, attr->step, attr->sigma);
#else
  if (attr->itercnt == 0) {
    c_print("                  Objective                          Residual\n");
    c_print("Iter       Primal           Dual             Primal            Dual\n");
  }
  c_print("%2d    %+10.8e  %+10.8e  %+10.8e  %+10.8e\n", (int)attr->itercnt, attr->pobj,
          attr->dobj, attr->pres, attr->dres);
#endif

#ifdef MATLAB_MEX_FILE
  mexEvalString("pause(0.0001);");
#endif
}

/* Update solution */
static void solve_updatesol(csocp_prob *prob, c_int ifRestore) {
  c_real *srcX = NULL, *srcY = NULL, *srcZ = NULL, *srcS = NULL;
  c_real *dstX = NULL, *dstY = NULL, *dstZ = NULL, *dstS = NULL;
  c_attr *srcAttr = NULL, *dstAttr = NULL;

  if (ifRestore) {
    srcX = prob->best_x;
    srcY = prob->best_y;
    srcZ = prob->best_z;
    srcS = prob->best_s;

    dstX = prob->x;
    dstY = prob->y;
    dstZ = prob->z;
    dstS = prob->s;

    srcAttr = prob->best_attr;
    dstAttr = prob->attr;
  } else {
    srcX = prob->x;
    srcY = prob->y;
    srcZ = prob->z;
    srcS = prob->s;

    dstX = prob->best_x;
    dstY = prob->best_y;
    dstZ = prob->best_z;
    dstS = prob->best_s;

    srcAttr = prob->attr;
    dstAttr = prob->best_attr;
  }

  /* Update primal/dual solution */
  c_math_copydbl(prob->n, srcX, dstX, 0);
  c_math_copydbl(prob->p, srcY, dstY, 0);
  c_math_copydbl(prob->m, srcZ, dstZ, 0);
  c_math_copydbl(prob->m, srcS, dstS, 0);

  /* Update attributes */
  c_attr_copy(srcAttr, dstAttr);

  return;
}

/* Compares solA and solB, return 1 if solA is better, 0 otherwize */
static c_int solve_comparesol(c_attr *solA, c_attr *solB) {
  c_int cond_gap = 0, cond_mu = 0, cond_kap = 0;
  c_int cond_pres = 0, cond_dres = 0, cond_infres = 0;

  cond_gap = solA->absgap > 0 && solB->absgap > 0 && solA->absgap < solB->absgap;
  cond_mu = solA->mu > 0 && solA->mu < solB->mu;
  cond_kap = solA->kap_t > 0 && solA->kap_t < solB->kap_t;

  cond_pres = solA->pres > 0 && solA->pres < solB->pres;
  cond_dres = solA->dres > 0 && solA->dres < solB->dres;
  if (solA->pinfres != CSOCP_INFINITY) {
    cond_infres = solA->pinfres > 0 && solA->pinfres < solB->pinfres;
  }

  if (solA->pinfres != CSOCP_INFINITY && solA->kap_t > 1) {
    if (solB->pinfres != CSOCP_INFINITY) {
      /* Both pinfres are valid */
      if (cond_gap && cond_infres && cond_mu) {
        return 1;
      } else {
        return 0;
      }
    } else {
      /* Only pinfres of solA is valid */
      if (cond_gap && cond_mu) {
        return 1;
      } else {
        return 0;
      }
    }
  } else {
    /* Both pinfres are invalid */
    if (cond_gap && cond_pres && cond_dres && cond_kap && cond_mu) {
      return 1;
    } else {
      return 0;
    }
  }
}

/* Post-solve solution */
static void solve_postsol(csocp_prob *prob) {
  c_attr *attr = prob->attr;
#ifdef CSOCP_WITH_C6678
  c_real one_tau = c_math_divd(1.0, attr->tau);
#else
  c_real one_tau = 1.0 / attr->tau;
#endif

  /* Unscaling solution */
  c_math_scalemul(prob->n, one_tau, prob->x);
  c_math_scalemul(prob->p, one_tau, prob->y);
  c_math_scalemul(prob->m, one_tau, prob->z);
  c_math_scalemul(prob->m, one_tau, prob->s);

  c_math_div(prob->n, prob->facX, prob->x);
  c_math_div(prob->p, prob->facA, prob->y);
  c_math_div(prob->m, prob->facG, prob->z);
  c_math_mul(prob->m, prob->facG, prob->s);

  return;
}

/* Check if solution is optimal */
static c_int solve_isopt(csocp_prob *prob, c_real abstol, c_real feastol,
                         c_real reltol) {
  c_attr *attr = prob->attr;
  c_int cond_lvl = 0, cond_res = 0, cond_gap = 0;

  cond_lvl = attr->cx < 0 || attr->by + attr->hz <= abstol;
  cond_res = attr->pres < feastol && attr->dres < feastol;
  cond_gap = attr->absgap < abstol || attr->relgap < reltol;

  return cond_lvl && cond_res && cond_gap;
}

/* Check if solution is primal infeasible */
static c_int solve_ispinf(csocp_prob *prob, c_real feastol) {
  c_attr *attr = prob->attr;
  c_param *param = prob->param;
  c_int cond_res = 0, cond_tau = 0, cond_mix = 0;

  cond_res = attr->pinfres != CSOCP_INFINITY && attr->pinfres < feastol;
  cond_tau = attr->tau < attr->kap;
  cond_mix = attr->tau < param->feastol && attr->kap < param->feastol &&
             attr->pinfres < param->feastol;

  return ((cond_res && cond_tau) || cond_mix);
}

/* Check if solution is dual infeasible */
static c_int solve_isdinf(csocp_prob *prob, c_real feastol) {
  c_attr *attr = prob->attr;
  c_int cond_res = 0, cond_tau = 0;

  cond_res = attr->dinfres != CSOCP_INFINITY && attr->dinfres < feastol;
  cond_tau = attr->tau < attr->kap;

  return cond_res && cond_tau;
}

/* Check status by solution */
static c_int solve_checksol(csocp_prob *prob, c_int ifrelax) {
  c_attr *attr = prob->attr;
  c_param *param = prob->param;

  c_real feastol;
  c_real abstol;
  c_real reltol;

  if (ifrelax == 0) {
    feastol = param->feastol;
    abstol = param->abstol;
    reltol = param->reltol;
  } else {
    feastol = param->feastol_rlx;
    abstol = param->abstol_rlx;
    reltol = param->reltol_rlx;
  }

  if (solve_isopt(prob, abstol, feastol, reltol)) {
    c_print("\nOptimal solution found.");
    attr->ispinf = 0;
    attr->isdinf = 0;
    attr->status = CSOCP_OPTIMAL;
    return 0;
  } else if (solve_ispinf(prob, feastol)) {
    c_print("\nPrimal Infeasible.");
    attr->ispinf = 1;
    attr->isdinf = 0;
    attr->status = CSOCP_PRIMALINF;
    return 0;
  } else if (solve_isdinf(prob, feastol)) {
    c_print("\nDual infeasible.");
    attr->ispinf = 0;
    attr->isdinf = 1;
    attr->status = CSOCP_DUALINF;
    return 0;
  } else {
    attr->status = CSOCP_UNFINISHED;
    return -1;
  }

  return 0;
}

/* Solve problem */
c_int CSOCP_solve(csocp_prob *prob) {
  c_int ifbreak = 0;
  c_real solvebeg = 0.0;
  c_real res = 0.0;
  c_real old_pres = 0.0;

  c_attr *attr = prob->attr;
  c_param *param = prob->param;

  /* Start timer for solving process */
  solvebeg = c_timer_elapsed();

  /* Register interrupt handler */
  c_interrupt_register();

  /* Compute start point */
  if (c_solver_startpoint(prob)) {
    attr->status = CSOCP_ERROR;
    return 0;
  }

  /* Main interior point method loop */
  for (attr->itercnt = 0; attr->itercnt < param->maxit; attr->itercnt++) {
    /* Update iterative information */
    c_iter_updateattr(prob);

    /*
     * Backtrack if:
     *   1. Iterate is bad such that pres was increased by a factor of 'safeguard'
     *   2. The 'absgap' is negative
     *
     * If 'safeguard' is activated, the solver uses relax tolerances, if not even
     * relaxed tolerances are satisfied, report 'numerical'.
     */
    if (attr->itercnt > 0 &&
        (attr->pres > param->rlxfactor * old_pres || attr->absgap < 0.0)) {
      c_print("Backtrack solving step\n");

      /* Restore best iterate */
      solve_updatesol(prob, 1);

      /* Check with relaxed tolerances */
      if (solve_checksol(prob, 1) != 0) {
        attr->status = CSOCP_NUMERICAL;
      }
      break;
    }
    /* Store old primal residual */
    old_pres = attr->pres;

    /* Print solving log */
    solve_printlog(prob);

    /* Check for interrupt action */
    ifbreak = c_interrupt_check();

    /* Check solution with accurate tolerance */
    if (solve_checksol(prob, 0) != 0) {
      if (attr->itercnt > 0 && attr->step == param->minstep * param->gamma) {
        /* Small line search step */
        c_print("Numerical problem.\n");

        /* Restore best candidate if necessary */
        solve_updatesol(prob, 1);

        /* Check with relaxed tolerances */
        if (solve_checksol(prob, 1) != 0) {
          attr->status = CSOCP_NUMERICAL;
        }
        break;
      } else if (attr->itercnt == param->maxit - 1) {
        /* Maximal iteration limit */
        c_print("Iteration limit.\n");

        /* Restore best candidate if necessary */
        if (solve_comparesol(attr, prob->best_attr) == 0) {
          solve_updatesol(prob, 1);
        }

        /* Check with relaxed tolerances */
        if (solve_checksol(prob, 1) != 0) {
          attr->status = CSOCP_ITERLIMIT;
        }
        break;
      } else if (ifbreak) {
        /* User interrupt */
        c_print("User interrupted.\n");

        /* Restore best candidate if necessary */
        if (solve_comparesol(attr, prob->best_attr)) {
          solve_updatesol(prob, 1);
        }

        /* Check with relaxed tolerances */
        if (solve_checksol(prob, 1) != 0) {
          attr->status = CSOCP_INTERRUPTED;
        }
        break;
      } else if (attr->pobj == CSOCP_INFINITY) {
        /* Math error */
        c_print("Math error.\n");

        /* Restore best candidate if necessary */
        if (solve_comparesol(attr, prob->best_attr)) {
          solve_updatesol(prob, 1);
        }

        /* Unwanted situation */
        attr->status = CSOCP_ERROR;
        break;
      }
    } else {
      break;
    }

    /* Check if current solution is the best candidate */
    if (attr->itercnt == 0 || solve_comparesol(attr, prob->best_attr)) {
      solve_updatesol(prob, 0);
    }

    /* Update scalings of cones */
    if (c_cone_updatescale(prob->cone, prob->s, prob->z, prob->lambda, prob->buffer) !=
        CONE_OK) {
      /* Restore best candidate */
      solve_updatesol(prob, 1);

      /* Check with relaxed tolerances */
      if (solve_checksol(prob, 1) != 0) {
        attr->status = CSOCP_ERROR;
      }
      break;
    }

    /* Update KKT system */
    if (c_iter_updatekkt(prob) != 0) {
      attr->status = CSOCP_ERROR;
      break;
    }

    /* Update the predictor step */
    c_solver_predictor(prob, &res);

    /* Update centering parameter */
    c_solver_centering(prob);

    /* Update the corrector step */
    c_solver_corrector(prob, res);
  }

  /* Post-solve solution by unsetting scalings */
  solve_postsol(prob);

  /* Unscaling problem */
  c_scaling_unset(prob);

  /* Remove interrupt handler */
  c_interrupt_reset();

  /* Get elapsed solving time */
  attr->solve_time = c_timer_elapsed() - solvebeg;

  c_print("\nElapsed time: %.4fs (maxres: %.6e, relgap: %.6e, absgap: %.6e)\n",
          attr->load_time + attr->solve_time, CSOCP_MAX(attr->pres, attr->dres),
          attr->relgap, attr->absgap);

  return 0;
}

/* Get solution */
void CSOCP_getsolution(csocp_prob *prob, /* Pointer to problem */
                       c_real *x,        /* Primal solution */
                       c_real *y,        /* Dual solution of equalties */
                       c_real *s,        /* Slack of inequalities */
                       c_real *z         /* Dual of inequalities */
) {
  if (x != NULL) {
    c_math_copydbl(prob->n, prob->x, x, 0);
  }
  if (y != NULL) {
    c_math_copydbl(prob->p, prob->y, y, 0);
  }
  if (s != NULL) {
    c_math_copydbl(prob->m, prob->s, s, 0);
  }
  if (z != NULL) {
    c_math_copydbl(prob->m, prob->z, z, 0);
  }
  return;
}

/* Set double parameter */
void CSOCP_setdblparam(csocp_prob *prob, /* Pointer to problem */
                       c_int dblparam,   /* Double paramter */
                       c_real newval     /* New value */
) {
  c_param_setdbl(prob->param, dblparam, newval);
}

/* Set integer parameter */
void CSOCP_setintparam(csocp_prob *prob, /* Pointer to problem */
                       c_int intparam,   /* Integer parameter */
                       c_int newval      /* New value */
) {
  c_param_setint(prob->param, intparam, newval);
}

/* Get double parameter */
c_real CSOCP_getdblparam(csocp_prob *prob, /* Pointer to problem */
                         c_int dblparam    /* Double paramter */
) {
  return c_param_getdbl(prob->param, dblparam);
}

/* Get integer parameter */
c_int CSOCP_getintparam(csocp_prob *prob, /* Pointer to problem */
                        c_int intparam    /* Integer parameter */
) {
  return c_param_getint(prob->param, intparam);
}

/* Get double attribute */
c_real CSOCP_getdblattr(csocp_prob *prob, /* Pointer to problem */
                        c_int dblattr     /* Double attribute */
) {
  return c_attr_getdbl(prob->attr, dblattr);
}

/* Get integer attribute */
c_int CSOCP_getintattr(csocp_prob *prob, /* Pointer to problem */
                       c_int intattr     /* Integer attribute */
) {
  return c_attr_getint(prob->attr, intattr);
}

/* Free problem */
void CSOCP_free(csocp_prob **p_prob) {
  if (p_prob != NULL && *p_prob != NULL) {
#ifdef CSOCP_WITH_DEBUG
    if ((*p_prob)->dbgfile != NULL) {
      fclose((*p_prob)->dbgfile);
    }
#endif

#ifdef CSOCP_WITH_COPYPROB
    c_spmat_free(&(*p_prob)->A);
    c_spmat_free(&(*p_prob)->G);
    CSOCP_FREE((*p_prob)->c);
    CSOCP_FREE((*p_prob)->b);
    CSOCP_FREE((*p_prob)->h);
#else
    CSOCP_FREE((*p_prob)->A);
    CSOCP_FREE((*p_prob)->G);
#endif

    c_buffer_free(&(*p_prob)->buffer);
    c_cone_free(&(*p_prob)->cone);
    c_kkt_free(&(*p_prob)->kkt);

    CSOCP_FREE((*p_prob)->param);

    CSOCP_FREE((*p_prob)->facX);
    CSOCP_FREE((*p_prob)->facA);
    CSOCP_FREE((*p_prob)->facG);

    CSOCP_FREE((*p_prob)->rx);
    CSOCP_FREE((*p_prob)->ry);
    CSOCP_FREE((*p_prob)->rz);

    CSOCP_FREE((*p_prob)->dx1);
    CSOCP_FREE((*p_prob)->dy1);
    CSOCP_FREE((*p_prob)->dz1);

    CSOCP_FREE((*p_prob)->dx2);
    CSOCP_FREE((*p_prob)->dy2);
    CSOCP_FREE((*p_prob)->dz2);

    CSOCP_FREE((*p_prob)->dsaff);
    CSOCP_FREE((*p_prob)->dzaff);
    CSOCP_FREE((*p_prob)->dsaff_w);
    CSOCP_FREE((*p_prob)->dzaff_w);

    CSOCP_FREE((*p_prob)->lambda);

    CSOCP_FREE((*p_prob)->x);
    CSOCP_FREE((*p_prob)->y);
    CSOCP_FREE((*p_prob)->s);
    CSOCP_FREE((*p_prob)->z);
    CSOCP_FREE((*p_prob)->attr);

    CSOCP_FREE((*p_prob)->best_x);
    CSOCP_FREE((*p_prob)->best_y);
    CSOCP_FREE((*p_prob)->best_s);
    CSOCP_FREE((*p_prob)->best_z);
    CSOCP_FREE((*p_prob)->best_attr);

    CSOCP_FREE(*p_prob);
  }
}
