#ifndef __CSOCP_H__
#define __CSOCP_H__

#ifdef __cplusplus
extern "C" {
#endif

#include "csocp_config.h"

/* Solving status */
#define CSOCP_UNSTARTED 0   /* Solving unstarted */
#define CSOCP_OPTIMAL 1     /* Found optimal solution */
#define CSOCP_PRIMALINF 2   /* Primal infeasible */
#define CSOCP_DUALINF 3     /* Dual infeasible */
#define CSOCP_ITERLIMIT 4   /* Maximum number of iterations reached */
#define CSOCP_NUMERICAL 5   /* Search direction unreliable */
#define CSOCP_INTERRUPTED 6 /* Solver interrupted */
#define CSOCP_UNFINISHED 7  /* Solve not finished */
#define CSOCP_ERROR 8       /* Fatal error in solver */

/* Double parameters */
#define CSOCP_DBLPARAM_FEASTOL 1    /* Primal/dual feasibility tolerance */
#define CSOCP_DBLPARAM_ABSTOL 2     /* Absolute tolerance */
#define CSOCP_DBLPARAM_RELTOL 3     /* Relative tolerance */
#define CSOCP_DBLPARAM_FEASTOLRLX 4 /* Relaxed primal/dual feasibility tolerance */
#define CSOCP_DBLPARAM_ABSTOLRLX 5  /* Relaxed absolute tolerance */
#define CSOCP_DBLPARAM_RELTOLRLX 6  /* Relaxed relative tolerance */
#define CSOCP_DBLPARAM_SCALETOL 7   /* Scaling tolerance */

/* Integer parameters */
#define CSOCP_INTPARAM_ITERLIMIT 1   /* Maximal number of iterations */
#define CSOCP_INTPARAM_SCALEMETHOD 2 /* Scaling method */
#define CSOCP_INTPARAM_SCALEITER 3   /* Iteration limit of scaling method */
#define CSOCP_INTPARAM_MUSTRATEGY 4  /* Mu update strategy */

/* Double attributes */
#define CSOCP_DBLATTR_PRIMALOBJ 1    /* Primal objective value */
#define CSOCP_DBLATTR_DUALOBJ 2      /* Dual objective value */
#define CSOCP_DBLATTR_PRIMALRES 3    /* Primal residual */
#define CSOCP_DBLATTR_DUALRES 4      /* Dual residual */
#define CSOCP_DBLATTR_PRIMALINFRES 5 /* Primal residual of infeasibilities */
#define CSOCP_DBLATTR_DUALINFRES 6   /* Dual residual of infeasibilities */
#define CSOCP_DBLATTR_ABSGAP 7       /* Absolute gap */
#define CSOCP_DBLATTR_RELGAP 8       /* Relative gap */
#define CSOCP_DBLATTR_LOADTIME 9     /* Loading time */
#define CSOCP_DBLATTR_SOLVETIME 10   /* Solving time */

/* Integer attributes */
#define CSOCP_INTATTR_STATUS 1    /* Solving status */
#define CSOCP_INTATTR_ISPINF 2    /* Is primal infeasible */
#define CSOCP_INTATTR_ISDINF 3    /* Is dual infeasible */
#define CSOCP_INTATTR_ITERCNT 4   /* Iteration counts */
#define CSOCP_INTATTR_NUMVAR 5    /* Number of variables */
#define CSOCP_INTATTR_NUMINEQ 6   /* Number of inequalities */
#define CSOCP_INTATTR_NUMEQ 7     /* Numbr of equalities */
#define CSOCP_INTATTR_NUMLPCONE 8 /* Number of LP cones */
#define CSOCP_INTATTR_NUMSOCONE 9 /* Number of SOC cones */
#define CSOCP_INTATTR_NUMNNZA 10  /* Number of nonzeros in A */
#define CSOCP_INTATTR_NUMNNZG 11  /* Number of nonzeros in G */

/* Create problem */
c_int CSOCP_create(csocp_prob **p_prob);

/* Set up SOCP problem */
c_int CSOCP_loadsocp(csocp_prob *prob, /* The CSOCP problem */
                     c_int n,          /* Number of variables */
                     c_int m,          /* Number of inequalities */
                     c_int p,          /* Number of equalities */
                     c_int l,          /* Number of positive orthant */
                     c_int ncones,     /* Number of SOC cones */
                     c_int *q,         /* Dimension for each SOC cone */
                     c_real *Gpr,      /* Element pointer of inequality matrix */
                     c_int *Gjc,       /* Begin pointer of inequality matrix */
                     c_int *Gir,       /* Index pointer of inequality matrix */
                     c_real *Apr,      /* Element pointer of equality matrix */
                     c_int *Ajc,       /* Begin pointer of equality matrix */
                     c_int *Air,       /* Index pointer of equality matrix */
                     c_real *c,        /* Objective costs */
                     c_real *h,        /* RHS of cones */
                     c_real *b         /* RHS of equalities */
);

/* Solve problem */
c_int CSOCP_solve(csocp_prob *prob /* Pointer to problem */
);

/* Get solution */
void CSOCP_getsolution(csocp_prob *prob, /* Pointer to problem */
                       c_real *x,        /* Primal solution */
                       c_real *y,        /* Dual solution of equalties */
                       c_real *s,        /* Slack of inequalities */
                       c_real *z         /* Dual of inequalities */
);

/* Set double parameter */
void CSOCP_setdblparam(csocp_prob *prob, /* Pointer to problem */
                       c_int dblparam,   /* Double paramter */
                       c_real newval     /* New value */
);

/* Set integer parameter */
void CSOCP_setintparam(csocp_prob *prob, /* Pointer to problem */
                       c_int intparam,   /* Integer parameter */
                       c_int newval      /* New value */
);

/* Get double parameter */
c_real CSOCP_getdblparam(csocp_prob *prob, /* Pointer to problem */
                         c_int dblparam    /* Double paramter */
);

/* Get integer parameter */
c_int CSOCP_getintparam(csocp_prob *prob, /* Pointer to problem */
                        c_int intparam    /* Integer parameter */
);

/* Get double attribute */
c_real CSOCP_getdblattr(csocp_prob *prob, /* Pointer to problem */
                        c_int dblattr     /* Double attribute */
);

/* Get integer attribute */
c_int CSOCP_getintattr(csocp_prob *prob, /* Pointer to problem */
                       c_int intattr     /* Integer attribute */
);

/* Free problem */
void CSOCP_free(csocp_prob **p_prob /* Pointer to problem */
);

#ifdef __cplusplus
}
#endif

#endif
