/**
 * This file implements attribute module for the CSOCP solver
 */

#include "csocp.h"

#include "csocp_attr.h"
#include "csocp_config.h"

#include "struct_attr.h"

/* Reset attributes to default values */
void c_attr_reset(c_attr *attr /* Input attribute structure */
) {
  /* Public attributes */
  attr->status = CSOCP_UNSTARTED;
  attr->ispinf = 0;
  attr->isdinf = 0;
  attr->itercnt = 0;

  attr->pobj = CSOCP_INFINITY;
  attr->dobj = CSOCP_INFINITY;
  attr->pres = CSOCP_INFINITY;
  attr->dres = CSOCP_INFINITY;
  attr->pinfres = CSOCP_INFINITY;
  attr->dinfres = CSOCP_INFINITY;
  attr->absgap = CSOCP_INFINITY;
  attr->relgap = CSOCP_INFINITY;

  /* Statistics */
  attr->n = 0;
  attr->m = 0;
  attr->p = 0;
  attr->l = 0;
  attr->ncones = 0;
  attr->nconedim = 0;
  attr->nnzA = 0;
  attr->nnzG = 0;
  attr->nK = 0;
  attr->nnzK = 0;
  attr->nnzL = 0;
  attr->nTmpI = 0;
  attr->nTmpD = 0;

  attr->load_time = 0;
  attr->solve_time = 0;
#ifdef CSOCP_WITH_STATS
  attr->total_mem = 0;
  attr->cone_mem = 0;
  attr->kkt_mem = 0;
  attr->solver_mem = 0;
  attr->order_mem = 0;
  attr->tmp_mem = 0;
  attr->free_mem = 0;
#endif

  /* Private attributes */
  attr->sigma = 0;
  attr->mu = 0;

  attr->kap_t = 0;
  attr->kap = 0;
  attr->tau = 0;
  attr->kapaff = 0;
  attr->tauaff = 0;
  attr->step = 0;
  attr->step_aff = 0;

  attr->nx = 0;
  attr->ny = 0;
  attr->nz = 0;
  attr->ns = 0;

  attr->cx = 0;
  attr->by = 0;
  attr->hz = 0;

  return;
}

/* Copy attributes */
void c_attr_copy(c_attr *src_attr, /* Input attribute structure */
                 c_attr *dst_attr  /* Output attribute structure */
) {
  /* Public attributes */
  dst_attr->itercnt = src_attr->itercnt;

  dst_attr->pobj = src_attr->pobj;
  dst_attr->dobj = src_attr->dobj;
  dst_attr->pres = src_attr->pres;
  dst_attr->dres = src_attr->dres;
  dst_attr->pinfres = src_attr->pinfres;
  dst_attr->dinfres = src_attr->dinfres;
  dst_attr->absgap = src_attr->absgap;
  dst_attr->relgap = src_attr->relgap;

  /* Private attributes */
  dst_attr->sigma = src_attr->sigma;
  dst_attr->mu = src_attr->mu;

  dst_attr->kap_t = src_attr->kap_t;
  dst_attr->kap = src_attr->kap;
  dst_attr->tau = src_attr->tau;
  dst_attr->kapaff = src_attr->kapaff;
  dst_attr->tauaff = src_attr->tauaff;
  dst_attr->step = src_attr->step;
  dst_attr->step_aff = src_attr->step_aff;

  dst_attr->cx = src_attr->cx;
  dst_attr->by = src_attr->by;
  dst_attr->hz = src_attr->hz;

  return;
}

/* Get double attribute */
c_real c_attr_getdbl(c_attr *attr, /* Input attribute structure */
                     c_int dblattr /* ID of double attribute */
) {
  c_real retval = 0.0;

  switch (dblattr) {
  case CSOCP_DBLATTR_PRIMALOBJ:
    retval = attr->pobj;
    break;
  case CSOCP_DBLATTR_DUALOBJ:
    retval = attr->dobj;
    break;
  case CSOCP_DBLATTR_PRIMALRES:
    retval = attr->pres;
    break;
  case CSOCP_DBLATTR_DUALRES:
    retval = attr->dres;
    break;
  case CSOCP_DBLATTR_PRIMALINFRES:
    retval = attr->pinfres;
    break;
  case CSOCP_DBLATTR_DUALINFRES:
    retval = attr->dinfres;
    break;
  case CSOCP_DBLATTR_ABSGAP:
    retval = attr->absgap;
    break;
  case CSOCP_DBLATTR_RELGAP:
    retval = attr->relgap;
    break;
  case CSOCP_DBLATTR_LOADTIME:
    retval = attr->load_time;
    break;
  case CSOCP_DBLATTR_SOLVETIME:
    retval = attr->solve_time;
    break;
  default:
    break;
  }

  return retval;
}

/* Get integer attribute */
c_int c_attr_getint(c_attr *attr, /* Input structure structure */
                    c_int intattr /* ID of integer attribute */
) {
  c_int retval = 0;

  switch (intattr) {
  case CSOCP_INTATTR_STATUS:
    retval = attr->status;
    break;
  case CSOCP_INTATTR_ISPINF:
    retval = attr->ispinf;
    break;
  case CSOCP_INTATTR_ISDINF:
    retval = attr->isdinf;
    break;
  case CSOCP_INTATTR_ITERCNT:
    retval = attr->itercnt;
    break;
  case CSOCP_INTATTR_NUMVAR:
    retval = attr->n;
    break;
  case CSOCP_INTATTR_NUMINEQ:
    retval = attr->m;
    break;
  case CSOCP_INTATTR_NUMEQ:
    retval = attr->p;
    break;
  case CSOCP_INTATTR_NUMLPCONE:
    retval = attr->l;
    break;
  case CSOCP_INTATTR_NUMSOCONE:
    retval = attr->ncones;
    break;
  case CSOCP_INTATTR_NUMNNZA:
    retval = attr->nnzA;
    break;
  case CSOCP_INTATTR_NUMNNZG:
    retval = attr->nnzG;
    break;
  default:
    break;
  }

  return retval;
}

/* Get memory requirement of attribute */
c_int c_attr_getmem() { return sizeof(c_attr); }
