/**
 * This file implements routines for cone in CSOCP solver
 *
 * This module serves as part of core module
 */

#include "csocp_cone.h"
#include "csocp_config.h"
#include "csocp_param.h"

#include "tool_buffer.h"
#include "tool_math.h"
#include "tool_memory.h"
#include "tool_print.h"
#include "tool_time.h"

#include "struct_cone.h"
#include "struct_param.h"

#include <stdlib.h>

/**
 * Creating and deleting methods
 */

/* Create LP cone */
static void cone_createlp(c_lp **p_lp, /* Pointer to LP cone */
                          c_int ndim   /* Dimension of LP cone */
) {
  c_lp *lp = NULL;

  /* Create the LP cone */
  CSOCP_INITZ(lp, 1);

  lp->ndim = ndim;
  if (ndim > 0) {
    CSOCP_INIT(lp->idxmap, ndim);
    CSOCP_INIT(lp->w, ndim);
    CSOCP_INIT(lp->v, ndim);
  }

  *p_lp = lp;

  return;
}

/* Free LP cone */
static void cone_freelp(c_lp **p_lp /* Pointer to LP cone */
) {
  if (p_lp != NULL && *p_lp != NULL) {
    if ((*p_lp)->ndim > 0) {
      CSOCP_FREE((*p_lp)->idxmap);
      CSOCP_FREE((*p_lp)->w);
      CSOCP_FREE((*p_lp)->v);
    }
    CSOCP_FREE(*p_lp);
  }
  return;
}

/* Create SOC cone */
static void cone_createsoc(c_soc **p_soc, /* Pointer to SOC cone */
                           c_int ndim     /* Dimension of SOC cone */
) {
  c_soc *soc = NULL;

  /* Create the SOC cone */
  CSOCP_INITZ(soc, 1);

  soc->ndim = ndim;
  if (ndim > 0) {
    CSOCP_INIT(soc->idxmap, ndim);
    CSOCP_INIT(soc->q, ndim);
  }

  *p_soc = soc;

  return;
}

/* Free SOC cone */
static void cone_freesoc(c_soc **p_soc /* Pointer to SOC cone */
) {
  if (p_soc != NULL && *p_soc != NULL) {
    CSOCP_FREE((*p_soc)->idxmap);
    CSOCP_FREE((*p_soc)->q);

    CSOCP_FREE(*p_soc);
  }
  return;
}

/* Creating cone */
void c_cone_create(c_cone **p_cone, /* Pointer to cone structure */
                   c_int nlpdim,    /* Dimension of LP cone */
                   c_int nsoc,      /* Number of SOC cones */
#ifndef CSOCP_WITH_DEBUG
                   c_int *socdim /* Dimensions of SOC cones */
#else
                   c_int *socdim, /* Dimensions of SOC cones */
                   FILE *dbgfile  /* Debug log file */
#endif
) {
  c_int i = 0;
  c_cone *cone = NULL;

  /* Create cone */
  CSOCP_INIT(cone, 1);

  /* Creating the LP cone */
  cone_createlp(&cone->lp, nlpdim);

  /* Creating the SOC cones */
  cone->nsoc = nsoc;

  CSOCP_INITZ(cone->soc, nsoc);
  for (i = 0; i < nsoc; ++i) {
    cone_createsoc(&cone->soc[i], (c_int)socdim[i]);
  }

#ifdef CSOCP_WITH_DEBUG
  /* Pass debug log file pointer */
  cone->dbgfile = dbgfile;
#endif

  *p_cone = cone;

  return;
}

/* Free cone */
void c_cone_free(c_cone **p_cone /* Pointer to cone structure */
) {
  c_int i = 0;
  if (p_cone != NULL && *p_cone != NULL) {
    /* Free the LP cone */
    cone_freelp(&(*p_cone)->lp);

    /* Free the SOC cones */
    for (i = 0; i < (*p_cone)->nsoc; ++i) {
      cone_freesoc(&(*p_cone)->soc[i]);
    }
    CSOCP_FREE((*p_cone)->soc);

    CSOCP_FREE(*p_cone);
  }
  return;
}

/**
 * Solving related methods
 */

/* Push variables to inside of cones: s = r + (1 + maxres) */
void c_cone_pushbnd(c_cone *cone,   /* Input cone structure */
                    c_real *srcval, /* Input values */
                    c_real *dstval, /* Output values */
                    c_param *param  /* Solver parameters */
) {
  c_int i = 0, iconeidx = 0;

  c_real res = 0.0, gamma = 0.0;
  c_real *dsrc = NULL, *ddst = NULL;

  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(cone->dbgfile, "[%s:%d]: Starting push variables to inside of cones\n",
           __FILE__, __LINE__);
#endif

  /**
   * Find maximal residual and store it in 'gamma'
   */
  gamma = -param->gamma;

  /* Firstly, find maximal residual in LP cone */
  for (i = 0; i < lp->ndim; ++i) {
    if (srcval[i] <= 0 && -srcval[i] > gamma) {
      gamma = -srcval[i];
    }
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(cone->dbgfile, "[%s:%d]: Maximal residual of LP cone: %.6e\n", __FILE__,
           __LINE__, gamma);
#endif

  /* Secondly, find maximal residual in SOC cones */
  iconeidx = lp->ndim;
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    dsrc = srcval + iconeidx;

    res = dsrc[0] - c_math_sqrt(c_math_dot(soc->ndim - 1, dsrc + 1, dsrc + 1));
    if (res <= 0 && -res > gamma) {
      gamma = -res;
    }

    iconeidx += soc->ndim;
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(cone->dbgfile, "[%s:%d]: Maximal residual of LP/SOC cone: %.6e\n", __FILE__,
           __LINE__, gamma);
#endif

  /**
   * Compute: dstval = srcval + (1 + gamma) * I
   *
   * where 'gamma' is the maximal residual of cones
   */
  gamma += 1.0;

  /* Firstly, update the LP cone */
  c_math_copydbl(lp->ndim, srcval, dstval, 0);
  c_math_add(lp->ndim, gamma, NULL, dstval);

  /* Secondly, update the SOC cones */
  iconeidx = lp->ndim;
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    dsrc = srcval + iconeidx;
    ddst = dstval + iconeidx;

    ddst[0] = dsrc[0] + gamma;
    c_math_copydbl(soc->ndim - 1, dsrc + 1, ddst + 1, 0);

    iconeidx += soc->ndim;
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(cone->dbgfile, "[%s:%d]: Finish push variables to inside of cones (%.4fs)\n",
           __FILE__, __LINE__, c_timer_elapsed() - timebeg);
#endif

  return;
}

/* Scale primal values: dstscal += w^2*srcprim */
void c_cone_scaleprimal(c_cone *cone,    /* Input cone structure */
                        c_real *srcprim, /* Input primal values */
                        c_real *dstscal  /* Output + Input dual values */
) {
  c_int i = 0, j = 0, iconeidx = 0;

  c_real res = 0.0;
  c_real *p1 = NULL, *p2 = NULL, *p3 = NULL, *p4 = NULL;
  c_real *d1 = NULL, *d2 = NULL, *d3 = NULL, *d4 = NULL;
  c_real *q1 = NULL;

  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(cone->dbgfile, "[%s:%d]: Starting scale primal values of cones\n", __FILE__,
           __LINE__);
#endif

  /**
   * Compute scalings
   *
   *           [D  v  u]
   *   eta^2 * [v' 1  0] = w^2
   *           [u  0 -1]
   */

  /* Firstly, compute the LP cone */
  for (i = 0; i < lp->ndim; ++i) {
    dstscal[i] += lp->v[i] * srcprim[i];
  }

  /* Secondly, compute the SOC cones */
  iconeidx = lp->ndim;
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    /* Pointer to SOC values */
    p1 = srcprim + iconeidx; /* Scalar */
    p2 = p1 + 1;             /* Vector */
    p3 = p1 + soc->ndim;     /* Scalar */
    p4 = p1 + soc->ndim + 1; /* Scalar */

    d1 = dstscal + iconeidx; /* Scalar */
    d2 = d1 + 1;             /* Vector */
    d3 = d1 + soc->ndim;     /* Scalar */
    d4 = d1 + soc->ndim + 1; /* Scalar */

    q1 = soc->q + 1;

    d1[0] += soc->eta_q * (soc->d1 * p1[0] + soc->u0 * p4[0]);

    res = soc->v1 * p3[0] + soc->u1 * p4[0];
    for (j = 0; j < soc->ndim - 1; ++j) {
      d2[j] += soc->eta_q * (p2[j] + res * q1[j]);
    }

    res = c_math_dot(soc->ndim - 1, q1, p2);

    d3[0] += soc->eta_q * (soc->v1 * res + p3[0]);
    d4[0] += soc->eta_q * (soc->u0 * p1[0] + soc->u1 * res - p4[0]);

    /* Update to next SOC */
    iconeidx += soc->ndim + 2;
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(cone->dbgfile, "[%s:%d]: Finish scale primal values of cones (%.4fs)\n",
           __FILE__, __LINE__, c_timer_elapsed() - timebeg);
#endif

  return;
}

/* Scale dual values: dstscal = w*srcdual */
void c_cone_scaledual(c_cone *cone,    /* Input cone structure */
                      c_real *srcdual, /* Input dual values */
                      c_real *dstscal  /* Output scalings */
) {
  c_int i = 0, j = 0, iconeidx = 0;

  c_real zeta = 0.0, factor = 0.0;
  c_real *dslack = NULL, *dscale = NULL;

  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(cone->dbgfile, "[%s:%d]: Starting scale dual values of cones\n", __FILE__,
           __LINE__);
#endif

  /**
   * Scale the cones
   *
   *   lambda = w*z
   */

  /* Firstly, scale the LP cone */
  for (i = 0; i < lp->ndim; ++i) {
    dstscal[i] = lp->w[i] * srcdual[i];
  }

  /* Secondly, scale the SOC cones */
  iconeidx = lp->ndim;
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    /* Pointer to first element for each SOC values */
    dslack = srcdual + iconeidx;
    dscale = dstscal + iconeidx;

    /* Compute scalings for each SOC */
    zeta = c_math_dot(soc->ndim - 1, soc->q + 1, dslack + 1);
    factor = dslack[0] + CSOCP_DIV(zeta, 1.0 + soc->q[0]);

    dscale[0] = soc->eta * (soc->q[0] * dslack[0] + zeta);
    for (j = 1; j < soc->ndim; ++j) {
      dscale[j] = soc->eta * (dslack[j] + factor * soc->q[j]);
    }

    /* Update to next SOC */
    iconeidx += soc->ndim;
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(cone->dbgfile, "[%s:%d]: Finish scale dual values of cones (%.4fs)\n",
           __FILE__, __LINE__, c_timer_elapsed() - timebeg);
#endif

  return;
}

/* Update scalings */
c_int c_cone_updatescale(c_cone *cone,    /* Input cone structure */
                         c_real *srcprim, /* Input primal values */
                         c_real *srcdual, /* Input dual values */
                         c_real *dstscal, /* Output scalings */
                         c_buffer *buffer /* Input buffer */
) {
  c_int i = 0, j = 0, iconeidx = 0;

  c_real *pslk = NULL, *dslk = NULL;
  c_real pres = 0.0, dres = 0.0;
  c_real pnorm = 0.0, dnorm = 0.0;

  c_real gamma = 0.0;
  c_real c = 0.0, d = 0.0, cquad = 0.0;

  c_real w = 0.0;
  c_real qq = 0.0, qone = 0.0;
  c_real u0 = 0.0, u02 = 0.0;
  c_real d1 = 0.0, u1 = 0.0, v1 = 0.0;

  c_real *sk = NULL;
  c_real *zk = NULL;

  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(cone->dbgfile, "[%s:%d]: Starting update scalings of cones\n", __FILE__,
           __LINE__);
#endif

  /**
   * Calculate scalings
   */

  /* Firstly, compute scalings of LP cone */
  for (i = 0; i < lp->ndim; ++i) {
    lp->v[i] = CSOCP_DIV(srcprim[i], srcdual[i]);
    lp->w[i] = c_math_sqrt(lp->v[i]);
  }

  /* Secondly, compute scalings of SOC cones */
  iconeidx = lp->ndim;
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    /* Pointer to first element for each SOC values */
    pslk = srcprim + iconeidx;
    dslk = srcdual + iconeidx;

    /* Check SOC status by its residual */
    pres = pslk[0] * pslk[0] - c_math_dot(soc->ndim - 1, pslk + 1, pslk + 1);
    dres = dslk[0] * dslk[0] - c_math_dot(soc->ndim - 1, dslk + 1, dslk + 1);
#ifdef CSOCP_WITH_DEBUG
    c_fprint(cone->dbgfile, "[%s:%d]: Primal-Dual SOC residual: %.6e, %.6e\n", __FILE__,
             __LINE__, pres, dres);
#endif
    if (pres <= 0 || dres <= 0) {
      return CONE_INVALID;
    }

    /* Normalize SOC variables */
    c_buffer_request_dbuf(buffer, sk, soc->ndim);
    c_buffer_request_dbuf(buffer, zk, soc->ndim);

    pnorm = c_math_sqrt(pres);
    dnorm = c_math_sqrt(dres);
    for (j = 0; j < soc->ndim; ++j) {
      sk[j] = CSOCP_DIV(pslk[j], pnorm);
    }
    for (j = 0; j < soc->ndim; ++j) {
      zk[j] = CSOCP_DIV(dslk[j], dnorm);
    }
    soc->eta_q = CSOCP_DIV(pnorm, dnorm);
    soc->eta = c_math_sqrt(soc->eta_q);

    /* Normalized Nesterov-Todd scaling point */
    gamma = (1.0 + c_math_dot(soc->ndim, sk, zk)) * 0.5;
    gamma = c_math_sqrt(gamma);
    gamma = CSOCP_DIV(0.5, gamma);

    soc->q[0] = gamma * (sk[0] + zk[0]);
    for (j = 1; j < soc->ndim; ++j) {
      soc->q[j] = gamma * (sk[j] - zk[j]);
    }

    w = c_math_dot(soc->ndim - 1, soc->q + 1, soc->q + 1);

    c_buffer_release_dbuf(buffer, soc->ndim); /* sk */
    c_buffer_release_dbuf(buffer, soc->ndim); /* dk */

    /* Compute u and v */
    qq = soc->q[0] * soc->q[0];
    qone = soc->q[0] + 1.0;
    c = qone + CSOCP_DIV(w, qone);
    cquad = c * c;

    d = 1.0 + CSOCP_DIV(2.0, qone) + CSOCP_DIV(w, qone * qone);

    d1 = 0.5 * (qq + w * (1.0 - CSOCP_DIV(cquad, 1.0 + w * d)));
    d1 = CSOCP_MAX(d1, 0.0);

    u02 = qq + w - d1;
    u0 = c_math_sqrt(u02);
    u02 = CSOCP_DIV(cquad, u02);
    if (u02 <= d) {
      return CONE_INVALID;
    }
    v1 = c_math_sqrt(u02 - d);
    u1 = c_math_sqrt(u02);

    soc->d1 = d1;
    soc->u0 = u0;
    soc->u1 = u1;
    soc->v1 = v1;

    /* Update to next SOC */
    iconeidx += soc->ndim;
  }

  /* Compute dstscal = w*srcdual */
  c_cone_scaledual(cone, srcdual, dstscal);

#ifdef CSOCP_WITH_DEBUG
  c_fprint(cone->dbgfile, "[%s:%d]: Finish update scalings of cones (%.4fs)\n",
           __FILE__, __LINE__, c_timer_elapsed() - timebeg);
#endif

  return CONE_OK;
}

/* Post-solve search directions */
void c_cone_postsolve(c_cone *cone,   /* Input cone structure */
                      c_int ncol,     /* Number of variables */
                      c_int neq,      /* Number of equalities */
                      c_int *invmap,  /* Inverse index map */
                      c_real *invval, /* Inverse values */
                      c_real *dx,     /* Search direction of x */
                      c_real *dy,     /* Search direction of y */
                      c_real *dz      /* Search direction of z */
) {
  c_int i = 0, iconeidx = 0;
  c_int *idxmap = invmap;

  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(cone->dbgfile, "\n");
  c_fprint(cone->dbgfile, "[%s:%d]: Starting post-solve solution of cones\n", __FILE__,
           __LINE__);
#endif

  /* Post-solve x part */
  c_math_copydblbyinvidx(ncol, idxmap, invval, dx);
  idxmap += ncol;

  /* Post-solve y part */
  c_math_copydblbyinvidx(neq, idxmap, invval, dy);
  idxmap += neq;

  /* Post-solve z part */
  c_math_copydblbyinvidx(lp->ndim, idxmap, invval, dz);
  idxmap += lp->ndim;

  iconeidx = lp->ndim;
  for (i = 0; i < cone->nsoc; i++) {
    soc = cone->soc[i];

    c_math_copydblbyinvidx(soc->ndim, idxmap, invval, dz + iconeidx);
    idxmap += soc->ndim + 2;

    iconeidx += soc->ndim;
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(cone->dbgfile, "[%s:%d]: Finish post-solve solution of cones (%.4fs)\n",
           __FILE__, __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(cone->dbgfile, "\n");
#endif

  return;
}

/**
 * Basic operator methods
 */

/* Conic product: w = u o v */
c_real c_cone_prod(c_cone *cone, /* Input cone structure */
                   c_real *srcU, /* Input U values */
                   c_real *srcV, /* Input V values */
                   c_real *dstW  /* Ouput W values */
) {
  c_int i = 0, j = 0, idx = 0, iconeidx = 0;

  c_real mu = 0.0;
  c_real *dleft = NULL, *dright = NULL;

  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(cone->dbgfile, "\n");
  c_fprint(cone->dbgfile, "[%s:%d]: Starting conic product of cones\n", __FILE__,
           __LINE__);
#endif

  /* Compute the LP cone */
  for (i = 0; i < lp->ndim; ++i, ++idx) {
    dstW[idx] = srcU[i] * srcV[i];
    mu += CSOCP_ABS(dstW[idx]);
  }

  /* Compute the SOC cone */
  iconeidx = lp->ndim;
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    /* Pointer to SOC values */
    dleft = srcU + iconeidx;
    dright = srcV + iconeidx;

    dstW[idx] = c_math_dot(soc->ndim, dleft, dright);
    mu += CSOCP_ABS(dstW[idx]);

    ++idx;
    for (j = 1; j < soc->ndim; ++j, ++idx) {
      dstW[idx] = dleft[0] * dright[j] + dright[0] * dleft[j];
    }

    /* Update to next SOC */
    iconeidx += soc->ndim;
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(cone->dbgfile, "[%s:%d]: Finish conic product of cones (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(cone->dbgfile, "\n");
#endif

  return mu;
}

/* Conic division: v = u \ w */
void c_cone_div(c_cone *cone, /* Input cone structure */
                c_real *srcU, /* Input U values */
                c_real *srcW, /* Input W values */
                c_real *dstV  /* Output V values */
) {
  c_int i = 0, j = 0, idx = 0, iconeidx = 0;

  c_real rho = 0.0, zeta = 0.0, factor = 0.0;
  c_real *dleft = NULL, *dright = NULL;

  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(cone->dbgfile, "\n");
  c_fprint(cone->dbgfile, "[%s:%d]: Starting conic division of cones\n", __FILE__,
           __LINE__);
#endif

  /* Compute the LP cone */
  for (i = 0; i < lp->ndim; ++i, ++idx) {
    dstV[idx] = CSOCP_DIV(srcW[i], srcU[i]);
  }

  /* Compute the SOC cone */
  iconeidx = lp->ndim;
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    /* Pointer to SOC values */
    dleft = srcU + iconeidx;
    dright = srcW + iconeidx;

    rho = dleft[0] * dleft[0] - c_math_dot(soc->ndim - 1, dleft + 1, dleft + 1);
    zeta = c_math_dot(soc->ndim - 1, dleft + 1, dright + 1);

    factor = CSOCP_DIV(zeta, dleft[0]) - dright[0];
    factor = CSOCP_DIV(factor, rho);

    dstV[idx++] = CSOCP_DIV(dleft[0] * dright[0] - zeta, rho);
    for (j = 1; j < soc->ndim; ++j, ++idx) {
      dstV[idx] = factor * dleft[j] + CSOCP_DIV(dright[j], dleft[0]);
    }

    /* Update to next SOC */
    iconeidx += soc->ndim;
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(cone->dbgfile, "[%s:%d]: Finish conic division of cones (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(cone->dbgfile, "\n");
#endif

  return;
}

/**
 * Auxiliary methods
 */

/* Get memory requirement of cone */
c_int c_cone_getmem(c_int lpdim,  /* Dimension of LP cone */
                    c_int nsoc,   /* Number of SOC cones */
                    c_int *socdim /* Dimensions of SOC cones */
) {
  c_int i = 0;
  c_int conesize = 0;

  /* Memory of cone structure */
  conesize += sizeof(c_lp *);         /* lp */
  conesize += sizeof(c_int);          /* nsoc */
  conesize += nsoc * sizeof(c_soc *); /* soc */

  /* Memory of LP cone */
  conesize = sizeof(c_int);           /* ndim */
  conesize += lpdim * sizeof(c_int);  /* idxmap */
  conesize += lpdim * sizeof(c_real); /* w */
  conesize += lpdim * sizeof(c_real); /* v */

  /* Memory of SOC cones */
  for (i = 0; i < nsoc; ++i) {
    conesize += sizeof(c_int);              /* ndim */
    conesize += socdim[i] * sizeof(c_int);  /* idxmap */
    conesize += socdim[i] * sizeof(c_real); /* q */
    conesize += sizeof(c_real);             /* eta */
    conesize += sizeof(c_real);             /* eta_q */
    conesize += sizeof(c_real);             /* d1 */
    conesize += sizeof(c_real);             /* u0 */
    conesize += sizeof(c_real);             /* u1 */
    conesize += sizeof(c_real);             /* v1 */
  }

  return conesize;
}
