/**
 * This file implements routines for cone in CSOCP solver
 *
 * This module serves as part of core module
 */

#ifndef __CSOCP_CONE_H__
#define __CSOCP_CONE_H__

#include "csocp_config.h"
#ifdef CSOCP_WITH_DEBUG
#include <stdio.h>
#endif

/* Cone status */
#define CONE_OK 0
#define CONE_INVALID 1

/**
 * Creating and deleting methods
 */

/* Creating cone */
void c_cone_create(c_cone **p_cone, /* Pointer to cone structure */
                   c_int nlpdim,    /* Dimension of LP cone */
                   c_int nsoc,      /* Number of SOC cones */
#ifndef CSOCP_WITH_DEBUG
                   c_int *socdim /* Dimensions of SOC cones */
#else
                   c_int *socdim, /* Dimensions of SOC cones */
                   FILE *dbgfile  /* Debug log file */
#endif
);

/* Free cone */
void c_cone_free(c_cone **p_cone /* Pointer to cone structure */
);

/**
 * Solving related methods
 */

/* Push variables to inside of cones */
void c_cone_pushbnd(c_cone *cone,   /* Input cone structure */
                    c_real *srcval, /* Input values */
                    c_real *dstval, /* Output values */
                    c_param *param  /* Solver parameters */
);

/* Scale primal values: dstscal += w^2*srcprim */
void c_cone_scaleprimal(c_cone *cone,    /* Input cone structure */
                        c_real *srcprim, /* Input primal values */
                        c_real *dstscal  /* Output + Input dual values */
);

/* Scale dual values: dstscal = w*srcdual */
void c_cone_scaledual(c_cone *cone,    /* Input cone structure */
                      c_real *srcdual, /* Input dual values */
                      c_real *dstscal  /* Output scalings */
);

/* Update scalings */
c_int c_cone_updatescale(c_cone *cone,    /* Input cone structure */
                         c_real *srcprim, /* Input primal values */
                         c_real *srcdual, /* Input dual values */
                         c_real *dstscal, /* Output scalings */
                         c_buffer *buffer /* Input buffer */
);

/* Post-solve search directions */
void c_cone_postsolve(c_cone *cone,   /* Input cone structure */
                      c_int ncol,     /* Number of variables */
                      c_int neq,      /* Number of equalities */
                      c_int *invmap,  /* Inverse index map */
                      c_real *invval, /* Inverse values */
                      c_real *dx,     /* Search direction of x */
                      c_real *dy,     /* Search direction of y */
                      c_real *dz      /* Search direction of z */
);

/**
 * Basic operator methods
 */

/* Conic product: w = u o v */
c_real c_cone_prod(c_cone *cone, /* Input cone structure */
                   c_real *srcU, /* Input U values */
                   c_real *srcV, /* Input V values */
                   c_real *dstW  /* Ouput W values */
);

/* Conic division: v = u \ w */
void c_cone_div(c_cone *cone, /* Input cone structure */
                c_real *srcU, /* Input U values */
                c_real *srcW, /* Input W values */
                c_real *dstV  /* Output V values */
);

/**
 * Auxiliary methods
 */

/* Get memory requirement of cone */
c_int c_cone_getmem(c_int lpdim,  /* Dimension of LP cone */
                    c_int nsoc,   /* Number of SOC cones */
                    c_int *socdim /* Dimensions of SOC cones */
);

#endif
