/**
 * This file implements iterative module of the CSOCP solver
 */

#include "csocp_iter.h"
#include "csocp_config.h"
#include "csocp_kkt.h"

#include "struct_attr.h"
#include "struct_kkt.h"
#include "struct_param.h"
#include "struct_prob.h"

#include "tool_math.h"
#include "tool_print.h"
#include "tool_spmat.h"
#include "tool_time.h"

/* Update attributes */
void c_iter_updateattr(csocp_prob *prob /* The CSOCP problem */
) {
  c_real nrx = 0.0, nry = 0.0, nrz = 0.0;
  c_real resx = 0.0, resy = 0.0, resz = 0.0;
  c_real one_tau = 0.0, tmpval = 0.0;

  c_attr *attr = prob->attr;
  c_param *param = prob->param;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting update attributes\n", __FILE__, __LINE__);
#endif

  /*
   * Update residuals
   */

  /* Compute: rx = -A'*y - G'*z */
  if (prob->p > 0) {
    c_spmat_mattvec(prob->A, prob->y, prob->rx, 1, 0);
    c_spmat_mattvec(prob->G, prob->z, prob->rx, 0, 0);
  } else {
    c_spmat_mattvec(prob->G, prob->z, prob->rx, 1, 0);
  }
  resx = c_math_norm2(prob->n, prob->rx);
  /* Compute: rx = rx - c.*tau */
  c_math_sub(prob->n, attr->tau, prob->c, prob->rx);

  /* Compute: ry = A*x */
  if (prob->p > 0) {
    c_spmat_matvec(prob->A, prob->x, prob->ry, 1, 1);
    resy = c_math_norm2(prob->p, prob->ry);
    /* Compute: ry = ry - b.*tau */
    c_math_sub(prob->p, attr->tau, prob->b, prob->ry);
  }

  /* Compute: rz = G*x + s */
  c_spmat_matvec(prob->G, prob->x, prob->rz, 1, 1);
  c_math_add(prob->m, 1.0, prob->s, prob->rz);
  resz = c_math_norm2(prob->m, prob->rz);
  /* Compute: rz = rz - h.*tau */
  c_math_sub(prob->m, attr->tau, prob->h, prob->rz);

  /* Compute: c'*x, b'*y, h'*z */
  attr->cx = c_math_dot(prob->n, prob->c, prob->x);
  if (prob->p > 0) {
    attr->by = c_math_dot(prob->p, prob->b, prob->y);
  }
  attr->hz = c_math_dot(prob->m, prob->h, prob->z);

  /* Compute norms */
  attr->nx = c_math_norm2(prob->n, prob->x);
  attr->ny = c_math_norm2(prob->p, prob->y);
  attr->nz = c_math_norm2(prob->m, prob->z);
  attr->ns = c_math_norm2(prob->m, prob->s);

  /*
   * Update statistics
   */

#ifdef CSOCP_WITH_C6678
  one_tau = c_math_divd(1.0, attr->tau);
#else
  one_tau = 1.0 / attr->tau;
#endif

  /* Update absolute gap: s'*z */
  attr->absgap = c_math_dot(prob->m, prob->s, prob->z);
  /* Update mu: (absgap + kap*tau) / (D + 1) */
#ifdef CSOCP_WITH_C6678
  attr->mu = c_math_divd(attr->absgap + attr->kap * attr->tau, prob->degree + 1.0);
#else
  attr->mu = (attr->absgap + attr->kap * attr->tau) / (prob->degree + 1.0);
#endif
  /* Update kap / tau */
  attr->kap_t = attr->kap * one_tau;

  /* Update primal objective */
  attr->pobj = attr->cx * one_tau;
  /* Update dual objective */
  attr->dobj = -(attr->hz + attr->by) * one_tau;

  /* Update relative gap */
  if (attr->pobj < 0) {
#ifdef CSOCP_WITH_C6678
    attr->relgap = -c_math_divd(attr->absgap, attr->pobj);
#else
    attr->relgap = -attr->absgap / attr->pobj;
#endif
  } else if (attr->dobj > 0) {
#ifdef CSOCP_WITH_C6678
    attr->relgap = c_math_divd(attr->absgap, attr->dobj);
#else
    attr->relgap = attr->absgap / attr->dobj;
#endif
  } else {
    attr->relgap = CSOCP_INFINITY;
  }

  /* Compute norm of residuals */
  nrx = c_math_norm2(prob->n, prob->rx);
  if (prob->p > 0) {
    nry = c_math_norm2(prob->p, prob->ry);
#ifdef CSOCP_WITH_C6678
    nry = c_math_divd(nry, CSOCP_MAX(attr->resy0 + attr->nx, 1.0));
#else
    nry /= CSOCP_MAX(prob->resy0 + attr->nx, 1.0);
#endif
  }
  nrz = c_math_norm2(prob->m, prob->rz);
#ifdef CSOCP_WITH_C6678
  nrz = c_math_divd(nrz, CSOCP_MAX(attr->resz0 + attr->nx + attr->ns, 1.0));
#else
  nrz /= CSOCP_MAX(prob->resz0 + attr->nx + attr->ns, 1.0);
#endif

  /* Update primal residual */
  attr->pres = CSOCP_MAX(nry, nrz) * one_tau;
  /* Update dual residual */
  attr->dres = nrx * one_tau;
#ifdef CSOCP_WITH_C6678
  attr->dres = c_math_divd(attr->dres, CSOCP_MAX(prob->resx0 + attr->ny + attr->nz, 1));
#else
  attr->dres /= CSOCP_MAX(prob->resx0 + attr->ny + attr->nz, 1.0);
#endif

  /* Update primal infeasibilities */
  tmpval = CSOCP_MAX(attr->ny + attr->nz, 1.0);
#ifdef CSOCP_WITH_C6678
  if (attr->hz + attr->by < -param->reltol * tmpval) {
    attr->pinfres = c_math_divd(resx, tmpval);
  }
#else
  if (attr->hz + attr->by < -param->reltol * tmpval) {
    attr->pinfres = resx / tmpval;
  }
#endif
  else {
    attr->pinfres = CSOCP_INFINITY;
  }
  /* Update dual infeasibilities */
  tmpval = CSOCP_MAX(attr->nx, 1.0);
#ifdef CSOCP_WITH_C6678
  if (attr->cx < -param->reltol * tmpval) {
    attr->dinfres = CSOCP_MAX(c_math_divd(resy, tmpval),
                              c_math_divd(resz, CSOCP_MAX(attr->nx + attr->ns, 1.0)));
  }
#else
  if (attr->cx < -param->reltol * tmpval) {
    attr->dinfres =
        CSOCP_MAX(resy / tmpval, resz / CSOCP_MAX(attr->nx + attr->ns, 1.0));
  }
#endif
  else {
    attr->dinfres = CSOCP_INFINITY;
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish update attributes (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif

  return;
}

/* Update KKT system */
c_int c_iter_updatekkt(csocp_prob *prob /* The CSOCP problem */
) {
  c_kkt *kkt = prob->kkt;

  c_param *param = prob->param;
  c_attr *attr = prob->attr;

  c_buffer *buffer = prob->buffer;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting new KKT update and solve\n", __FILE__,
           __LINE__);
#endif

  /* Update KKT matrix with scalings */
  c_kkt_update(kkt, prob->cone, param);

  /* Factor KKT matrix */
  c_kkt_factor(kkt, buffer, param, attr);

  /* Solve for RHS1, used in combined direction */
  c_kkt_solve(kkt, prob->n, prob->p, prob->m, prob->cone, prob->A, prob->G, kkt->rhs1,
              prob->dx1, prob->dy1, prob->dz1, 0, buffer, param, attr);

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish new KKT update and solve (%.4fs)\n",
           __FILE__, __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif

  return 0;
}
