/**
 * This file impelements KKT module of the CSOCP solver
 */

#include "csocp_kkt.h"
#include "csocp_cone.h"
#include "csocp_config.h"
#include "csocp_ldl.h"

#include "tool_buffer.h"
#include "tool_math.h"
#include "tool_memory.h"
#include "tool_print.h"
#include "tool_spmat.h"
#include "tool_time.h"

#include "struct_attr.h"
#include "struct_buffer.h"
#include "struct_cone.h"
#include "struct_kkt.h"
#include "struct_param.h"
#include "struct_spmat.h"

#include "amd.h"
#include "amd_internal.h"

#include <stdlib.h>

/**
 * Creating and deleting methods
 */

/* Creating KKT structure */
void c_kkt_create(c_kkt **p_kkt, /* Pointer to KKT structure */
                  c_int nK,      /* Number of columns in KKT matrix */
#ifndef CSOCP_WITH_DEBUG
                  c_int nnzK /* Number of elements in KKT matrix */
#else
                  c_int nnzK,   /* Number of elements in KKT matrix */
                  FILE *dbgfile /* Debug log file */
#endif
) {
  c_kkt *kkt = NULL;

  CSOCP_INITZ(kkt, 1);

  kkt->kmat = c_spmat_create(nK, nK, nnzK);
  CSOCP_INIT(kkt->D, nK);
  CSOCP_INIT(kkt->idxmap, nnzK);
  CSOCP_INIT(kkt->invmap, nK);
  CSOCP_INIT(kkt->etree, nK);
  CSOCP_INIT(kkt->sign, nK);
  CSOCP_INIT(kkt->rhs1, nK);
  CSOCP_INIT(kkt->rhs2, nK);

#ifdef CSOCP_WITH_DEBUG
  /* Pass debug log file pointer */
  kkt->dbgfile = dbgfile;
#endif

  *p_kkt = kkt;

  return;
}

/* Free KKT structure */
void c_kkt_free(c_kkt **p_kkt /* Pointer to KKT structure */
) {
  if (p_kkt != NULL && *p_kkt != NULL) {
    /* Free memory of KKT */
    c_spmat_free(&(*p_kkt)->kmat);
    c_spmat_free(&(*p_kkt)->L);
    CSOCP_FREE((*p_kkt)->D);
    CSOCP_FREE((*p_kkt)->idxmap);
    CSOCP_FREE((*p_kkt)->invmap);
    CSOCP_FREE((*p_kkt)->etree);
    CSOCP_FREE((*p_kkt)->sign);
    CSOCP_FREE((*p_kkt)->rhs1);
    CSOCP_FREE((*p_kkt)->rhs2);

    CSOCP_FREE(*p_kkt);
  }
  return;
}

/**
 * Solve related methods
 */

/* Initializes the KKT system */
void c_kkt_init(c_kkt *kkt,    /* Input KKT structure */
                c_int n,       /* Number of variables */
                c_int p,       /* Number of inequalities */
                c_cone *cone,  /* Input cone */
                c_real *c,     /* Objective costs */
                c_real *b,     /* RHS of equalities */
                c_real *h,     /* RHS of inequalities */
                c_param *param /* Solving parametrs */
) {
  c_int i = 0, iconeidx = 0;

  c_spmat *kmat = kkt->kmat;
  c_real *kpr = kmat->colElem;

  c_real *rhs1 = kkt->rhs1;
  c_real *rhs2 = kkt->rhs2;

  c_int *idxmap = kkt->idxmap;
  c_int *invmap = kkt->invmap;

  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(kkt->dbgfile, "\n");
  c_fprint(kkt->dbgfile, "[%s:%d]: Starting initialize KKT\n", __FILE__, __LINE__);
#endif

  /* Build KKT for LP cone */
  c_math_fillbyidx(lp->ndim, -1.0, kkt->idxmap, lp->idxmap, kpr);

  /* Build KKT for SOC cone */
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    /* Diagonal D */
    idxmap = kkt->idxmap;
    c_math_fillbyidx(soc->ndim, -1.0, idxmap, soc->idxmap, kpr);

    /* v */
    idxmap += soc->idxmap[soc->ndim - 1];
    c_math_fillbyidx(soc->ndim - 1, 0.0, idxmap + 1, NULL, kpr);
    kpr[idxmap[soc->ndim]] = -1.0;

    /* u */
    idxmap += soc->ndim + 1;
    c_math_fillbyidx(soc->ndim, 0.0, idxmap, NULL, kpr);
    kpr[idxmap[soc->ndim]] = 1.0;
  }

  /* Build RHS1: [0; b; h] */
  invmap = kkt->invmap;
  c_math_fillbyidx(n, 0.0, invmap, NULL, rhs1);
  invmap += n;
  c_math_copydblbyidx(p, invmap, b, rhs1, 0);
  invmap += p;
  c_math_copydblbyidx(lp->ndim, invmap, h, rhs1, 0);
  invmap += lp->ndim;

  iconeidx = lp->ndim;
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    c_math_copydblbyidx(soc->ndim, invmap, h + iconeidx, rhs1, 0);
    invmap += soc->ndim;

    rhs1[invmap[0]] = 0.0;
    rhs1[invmap[1]] = 0.0;
    invmap += 2;

    iconeidx += soc->ndim;
  }

  /* Build RHS2: [-c; 0; 0] */
  invmap = kkt->invmap;
  c_math_copydblbyidx(n, invmap, c, rhs2, 1);
  invmap += n;
  c_math_fillbyidx(kmat->nCol - n, 0.0, invmap, NULL, rhs2);

#ifdef CSOCP_WITH_DEBUG
  c_fprint(kkt->dbgfile, "[%s:%d]: Finish initialize KKT (%.4fs)\n", __FILE__, __LINE__,
           c_timer_elapsed() - timebeg);
  c_fprint(kkt->dbgfile, "\n");
#endif

  return;
}

/* Symbolic factorization of KKT system */
c_int c_kkt_symbolic(c_kkt *kkt,       /* Input KKT structure */
                     c_spmat *ku,      /* Input KKT matrix */
                     c_int *sign,      /* Ouput sign vector */
                     c_buffer *buffer, /* Input buffer */
                     c_attr *attr      /* Input/Output attribute structure */
) {
  c_int i = 0, status = 0;

  c_spmat *kmat = kkt->kmat;

  c_int *Ljc = NULL;
  c_int *Lir = NULL;
  c_real *Lpr = NULL;

  c_int *idx = NULL;
  c_int *lnz = NULL;
  c_int *flag = NULL;

  c_real param[AMD_CONTROL], info[AMD_INFO];

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(kkt->dbgfile, "\n");
  c_fprint(kkt->dbgfile, "[%s:%d]: Starting symbolic factorization of KKT\n", __FILE__,
           __LINE__);
#endif

  /* Calculate matrix ordering by AMD method */
  c_buffer_request_ibuf(buffer, idx, ku->nCol);

  AMD_defaults(param);
  status = AMD_order(ku->nCol, ku->colBeg, ku->colIdx, idx, param, info);
  if (status != AMD_OK) {
    return KKT_INVALID;
  }

#ifdef CSOCP_WITH_STATS
  /* Record ordering memory */
  attr->order_mem = info[AMD_MEMORY];
#endif

  /* Calculate index mapping of KKT matrix */
  c_math_invidx(ku->nCol, idx, kkt->invmap);
#ifdef CSOCP_WITH_DUMP
  c_math_dumpint(ku->nCol, idx, "order.txt");
  c_math_dumpint(ku->nCol, kkt->invmap, "order_inv.txt");
#endif
  c_buffer_release_ibuf(buffer, ku->nCol); /* idx */

  /* Permute the KKT matrix */
  c_spmat_permute(ku, kkt->invmap, kmat, kkt->idxmap, buffer);
#ifdef CSOCP_WITH_DUMP
  c_spmat_dump(kkt->kmat, "kmat.txt");
#endif

  /* Permute sign vector */
  c_math_copyintbyidx(ku->nCol, kkt->invmap, sign, kkt->sign, 0);
#ifdef CSOCP_WITH_DUMP
  c_math_dumpint(ku->nCol, sign, "sign.txt");
  c_math_dumpint(ku->nCol, kkt->sign, "sign_kkt.txt");
#endif

  /* Do symbolic factorization */
  c_buffer_request_ibuf(buffer, lnz, attr->nK);
  c_buffer_request_ibuf(buffer, flag, attr->nK);

  CSOCP_INIT(Ljc, ku->nCol + 1);
#ifdef CSOCP_WITH_DEBUG
  c_ldl_symbolic(kmat->nCol, kmat->colBeg, kmat->colIdx, Ljc, kkt->etree, lnz, flag,
                 kkt->dbgfile);
#else
  c_ldl_symbolic(kmat->nCol, kmat->colBeg, kmat->colIdx, Ljc, kkt->etree, lnz, flag);
#endif

  c_buffer_release_ibuf(buffer, attr->nK); /* lnz */
  c_buffer_release_ibuf(buffer, attr->nK); /* flag */

  /* Allocate memory for L */
  CSOCP_INIT(Lir, Ljc[ku->nCol]);
  CSOCP_INIT(Lpr, Ljc[ku->nCol]);
  kkt->L = c_spmat_generate(ku->nCol, ku->nCol, Ljc[ku->nCol], Ljc, Lir, Lpr);

  /* Recorde number of nonzeros in factor L */
  attr->nnzL = Ljc[ku->nCol];
#ifdef CSOCP_WITH_STATS
  attr->kkt_mem += c_spmat_getmem(ku->nCol, Ljc[ku->nCol]);
  attr->total_mem += c_spmat_getmem(ku->nCol, Ljc[ku->nCol]);
#endif

#ifdef CSOCP_WITH_DEBUG
  c_fprint(kkt->dbgfile, "[%s:%d]: Finish symbolic factorization of KKT (%.4fs)\n",
           __FILE__, __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(kkt->dbgfile, "\n");
#endif

  return KKT_OK;
}

/* Numerical factorization of KKT system */
void c_kkt_factor(c_kkt *kkt,       /* Input KKT structure */
                  c_buffer *buffer, /* Input buffer */
                  c_param *param,   /* Input parameters */
                  c_attr *attr      /* Input attributes */
) {
  c_spmat *kmat = kkt->kmat;
  c_spmat *lmat = kkt->L;

  c_int *lnz = NULL;
  c_int *pattern = NULL;
  c_int *flag = NULL;
  c_real *col = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(kkt->dbgfile, "\n");
  c_fprint(kkt->dbgfile, "[%s:%d]: Starting KKT factorization\n", __FILE__, __LINE__);
#endif

  /* Obtain temporary memory from buffer */
  c_buffer_request_ibuf(buffer, lnz, attr->nK);
  c_buffer_request_ibuf(buffer, pattern, attr->nK);
  c_buffer_request_ibuf(buffer, flag, attr->nK);
  c_buffer_request_dbuf(buffer, col, attr->nK);

  /* Numeric LDL factorization */
#ifdef CSOCP_WITH_DEBUG
  c_ldl_numeric(kmat->nCol, kmat->colBeg, kmat->colIdx, kmat->colElem, lmat->colBeg,
                kkt->etree, kkt->sign, param->epsilon, param->delta, lmat->colIdx,
                lmat->colElem, kkt->D, col, lnz, pattern, flag, kkt->dbgfile);
#else
  c_ldl_numeric(kmat->nCol, kmat->colBeg, kmat->colIdx, kmat->colElem, lmat->colBeg,
                kkt->etree, kkt->sign, param->epsilon, param->delta, lmat->colIdx,
                lmat->colElem, kkt->D, col, lnz, pattern, flag);
#endif

  /* Release buffer */
  c_buffer_release_ibuf(buffer, attr->nK); /* lnz */
  c_buffer_release_ibuf(buffer, attr->nK); /* pattern */
  c_buffer_release_ibuf(buffer, attr->nK); /* flag */
  c_buffer_release_dbuf(buffer, attr->nK); /* col */

#ifdef CSOCP_WITH_DEBUG
  c_fprint(kkt->dbgfile, "[%s:%d]: Finish KKT factorization (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(kkt->dbgfile, "\n");
#endif

  return;
}

/* Solve the KKT system */
void c_kkt_solve(c_kkt *kkt,       /* Input KKT structure */
                 c_int n,          /* Number of variables */
                 c_int p,          /* Number of inequalities */
                 c_int m,          /* Number of equalities */
                 c_cone *cone,     /* Input cone structure */
                 c_spmat *A,       /* Input equality matrix */
                 c_spmat *G,       /* Input inequality matrix */
                 c_real *rhs,      /* RHS values */
                 c_real *dx,       /* Search direction of x-part */
                 c_real *dy,       /* Search direction of y-part */
                 c_real *dz,       /* Search direction of z-part */
                 c_int isinit,     /* Is initial KKT solve? */
                 c_buffer *buffer, /* Input buffer */
                 c_param *param,   /* Input parameters */
                 c_attr *attr      /* Input attributes */
) {
  c_int i = 0, j = 0, k = 0;
  c_int r = 0, s = 0, t = 0, iter = 0;

  c_spmat *kmat = kkt->kmat;
  c_spmat *L = kkt->L;
  c_int *invmap = kkt->invmap;

  c_real *x = NULL;
  c_real *irx = NULL;
  c_real *b = NULL;
  c_real *irb = NULL;
  c_real *xbar = NULL;
  c_real *gx = NULL;

  c_real *ex = NULL;
  c_real *ey = NULL;
  c_real *ez = NULL;

  c_real nex = 0, ney = 0, nez = 0;
  c_real norm = 1.0 + c_math_norminf(attr->nK, rhs);

  c_real err = 0.0;
  c_real errold = CSOCP_INFINITY;
  c_real errtol = norm * param->linsystol;

  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(kkt->dbgfile, "\n");
  c_fprint(kkt->dbgfile, "[%s:%d]: Starting iterative refinement\n", __FILE__,
           __LINE__);
  if (p > 0) {
    c_fprint(kkt->dbgfile,
             "[%s:%d]: iter        ||ex||        ||ey||        ||ez|| (tol: %.6e)\n",
             __FILE__, __LINE__, errtol);
  } else {
    c_fprint(kkt->dbgfile, "[%s:%d]: iter        ||ex||        ||ez|| (tol: %.6e)\n",
             __FILE__, __LINE__, errtol);
  }
#endif

  /* Obtain temporary memory from buffer */
  c_buffer_request_dbuf(buffer, x, attr->nK);
  c_buffer_request_dbuf(buffer, irx, attr->nK);
  c_buffer_request_dbuf(buffer, b, attr->nK);
  c_buffer_request_dbuf(buffer, irb, attr->nK);
  c_buffer_request_dbuf(buffer, xbar, attr->nK);
  c_buffer_request_dbuf(buffer, gx, attr->nK);

  ex = b;
  ey = b + n;
  ez = b + n + p;

#ifdef CSOCP_WITH_DEBUG
  /* Forward solve: L*x = b */
  c_ldl_lsolve(attr->nK, rhs, L->colBeg, L->colIdx, L->colElem, x, kkt->dbgfile);
  /* Diagonal solve: D*x = b */
  c_ldl_dsolve(attr->nK, x, kkt->D, kkt->dbgfile);
  /* Backward solve: L'*x = b */
  c_ldl_ltsolve(attr->nK, x, L->colBeg, L->colIdx, L->colElem, kkt->dbgfile);
#else
  /* Forward solve: L*x = b */
  c_ldl_lsolve(attr->nK, rhs, L->colBeg, L->colIdx, L->colElem, x);
  /* Diagonal solve: D*x = b */
  c_ldl_dsolve(attr->nK, x, kkt->D);
  /* Backward solve: L'*x = b */
  c_ldl_ltsolve(attr->nK, x, L->colBeg, L->colIdx, L->colElem);
#endif

  /* Main iterative refinement loop */
  for (iter = 0; iter < param->iriter; ++iter) {
    /* Post-solve search direction */
    c_cone_postsolve(cone, n, p, invmap, x, dx, dy, dz);

    k = 0;
    r = 0;

    /* Error of dx: bx - A'*dy - G'*dz - delta * dx */
    for (i = 0; i < n; i++) {
      ex[i] = rhs[invmap[r++]] - param->deltastat * dx[i];
    }
    if (A) {
      c_spmat_mattvec(A, dy, ex, 0, 0);
    }
    c_spmat_mattvec(G, dz, ex, 0, 0);
    nex = c_math_norminf(n, ex);

    /* Error of dy: by - A*dx + delta*dy */
    if (p > 0) {
      for (i = 0; i < p; i++) {
        ey[i] = rhs[invmap[r++]] + param->deltastat * dy[i];
      }
      c_spmat_matvec(A, dx, ey, -1, 0);
      ney = c_math_norminf(p, ey);
    }

    k = 0;
    s = 0;
    t = 0;

    /* Error of dz: bz - G*dx + V*dz */
    c_spmat_matvec(G, dx, gx, 1, 1);
    for (i = 0; i < lp->ndim; i++) {
      ez[s++] = rhs[invmap[r++]] - gx[k++] + param->deltastat * dz[t++];
    }
    for (j = 0; j < cone->nsoc; j++) {
      soc = cone->soc[j];

      for (i = 0; i < soc->ndim - 1; i++) {
        ez[s++] = rhs[invmap[r++]] - gx[k++] + param->deltastat * dz[t++];
      }
      ez[s++] = rhs[invmap[r++]] - gx[k++] - param->deltastat * dz[t++];
      ez[s++] = 0;
      ez[s++] = 0;

      r += 2;
    }

    c_math_copydblbyinvidx(m + 2 * cone->nsoc, invmap + n + p, x, xbar);
    if (isinit == 0) {
      c_cone_scaleprimal(cone, xbar, ez);
    } else {
      c_math_add(m + 2 * cone->nsoc, 1.0, xbar, ez);
    }
    nez = c_math_norminf(m + 2 * cone->nsoc, ez);

    /* Maximal error in infinity norm */
    err = CSOCP_MAX(nex, nez);
    if (p > 0) {
      err = CSOCP_MAX(err, ney);
    }

#ifdef CSOCP_WITH_DEBUG
    if (p > 0) {
      c_fprint(kkt->dbgfile, "[%s:%d]: %2d    %.8e    %.8e    %.8e\n", __FILE__,
               __LINE__, iter, nex, ney, nez);
    } else {
      c_fprint(kkt->dbgfile, "[%s:%d]: %2d    %.8e    %.8e\n", __FILE__, __LINE__, iter,
               nex, ney);
    }
#endif

    /* Check if iterative refinement is valid, otherwise exit */
    if (iter > 0 && err > errold) {
      c_math_sub(attr->nK, 1.0, irx, x);
      break;
    }

    /* Check whether to do iterative refinement again */
    if (err < errtol || (iter > 0 && errold < param->irfact * err)) {
      break;
    }
    errold = err;

    /* Peturbate rhs */
    c_math_copydblbyidx(attr->nK, invmap, b, irb, 0);

#ifdef CSOCP_WITH_DEBUG
    /* Forward solve: L*x = b */
    c_ldl_lsolve(attr->nK, irb, L->colBeg, L->colIdx, L->colElem, irx, kkt->dbgfile);
    /* Diagonal solve: D*x = b */
    c_ldl_dsolve(attr->nK, irx, kkt->D, kkt->dbgfile);
    /* Backward solve: L'*x = b */
    c_ldl_ltsolve(attr->nK, irx, L->colBeg, L->colIdx, L->colElem, kkt->dbgfile);
#else
    /* Forward solve: L*x = b */
    c_ldl_lsolve(attr->nK, irb, L->colBeg, L->colIdx, L->colElem, irx);
    /* Diagonal solve: D*x = b */
    c_ldl_dsolve(attr->nK, irx, kkt->D);
    /* Backward solve: L'*x = b */
    c_ldl_ltsolve(attr->nK, irx, L->colBeg, L->colIdx, L->colElem);
#endif

    /* Refine solution */
    c_math_add(attr->nK, 1.0, irx, x);
  }

  /* Post-solve search direction */
  c_cone_postsolve(cone, n, p, invmap, x, dx, dy, dz);

  /* Release buffer */
  c_buffer_release_dbuf(buffer, attr->nK); /* x */
  c_buffer_release_dbuf(buffer, attr->nK); /* irx */
  c_buffer_release_dbuf(buffer, attr->nK); /* b */
  c_buffer_release_dbuf(buffer, attr->nK); /* irb */
  c_buffer_release_dbuf(buffer, attr->nK); /* xbar */
  c_buffer_release_dbuf(buffer, attr->nK); /* gx */

#ifdef CSOCP_WITH_DEBUG
  c_fprint(kkt->dbgfile, "[%s:%d]: Finish iterative refinement (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(kkt->dbgfile, "\n");
#endif

  return;
}

/* Update the KKT system */
void c_kkt_update(c_kkt *kkt,    /* Input KKT structure */
                  c_cone *cone,  /* Input cone structure */
                  c_param *param /* Solving parameters */
) {
  c_int i = 0, j = 0;
  c_real ddelta = 0.0;
  c_real *q1 = NULL;

  c_spmat *kmat = kkt->kmat;
  c_real *kpr = kmat->colElem;
  c_int *pidx = kkt->idxmap;

  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(kkt->dbgfile, "\n");
  c_fprint(kkt->dbgfile, "[%s:%d]: Starting update KKT matrix\n", __FILE__, __LINE__);
#endif

  /* Update the LP cone */
  for (i = 0; i < lp->ndim; ++i) {
    kpr[pidx[lp->idxmap[i]]] = -lp->v[i] - param->deltastat;
  }

  /* Update the SOC cone */
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    q1 = soc->q + 1;

    ddelta = soc->eta_q + param->deltastat;

    /* D */
    pidx = kkt->idxmap;
    kpr[pidx[soc->idxmap[0]]] = -soc->eta_q * soc->d1 - param->deltastat;
    c_math_fillbyidx(soc->ndim - 1, -ddelta, pidx, soc->idxmap + 1, kpr);

    /* v */
    pidx += soc->idxmap[soc->ndim - 1] + 1;
    for (j = 0; j < soc->ndim - 1; ++j) {
      kpr[pidx[j]] = -soc->eta_q * soc->v1 * q1[j];
    }
    kpr[pidx[soc->ndim - 1]] = -soc->eta_q;

    /* u */
    pidx += soc->ndim;
    kpr[pidx[0]] = -soc->eta_q * soc->u0;
    pidx += 1;
    for (j = 0; j < soc->ndim - 1; ++j) {
      kpr[pidx[j]] = -soc->eta_q * soc->u1 * q1[j];
    }
    kpr[pidx[soc->ndim - 1]] = ddelta;
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(kkt->dbgfile, "[%s:%d]: Finish update KKT matrix (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(kkt->dbgfile, "\n");
#endif

  return;
}

/**
 * Auxiliary methods
 */

/* Get memory requirement of KKT */
c_int c_kkt_getmem(c_int n,    /* Number of variables */
                   c_int m,    /* Number of inequalities */
                   c_int p,    /* Number of equalities */
                   c_int nnzA, /* Number of nonzeros of A */
                   c_int nnzG, /* Number of nonzeros of G */
                   c_int nK,   /* Number of columns in KKT */
                   c_int nnzK  /* Number of elements in KKT */
) {
  c_int kktsize = 0;

  /* Memory of KKT structure */
  kktsize += sizeof(c_kkt);

  /* Memory of KKT data */
  kktsize += c_spmat_getmem(nK, nnzK); /* kmat */
  /* L will be calculated in 'c_kkt_symbolic' */
  kktsize += nK * sizeof(c_real);  /* D */
  kktsize += nnzK * sizeof(c_int); /* idxmap */
  kktsize += nK * sizeof(c_int);   /* invmap */
  kktsize += nK * sizeof(c_int);   /* etree */
  kktsize += nK * sizeof(c_int);   /* sign */
  kktsize += nK * sizeof(c_real);  /* rhs1 */
  kktsize += nK * sizeof(c_real);  /* rhs2 */

  return kktsize;
}
