/**
 * This file impelements KKT module of the CSOCP solver
 */

#ifndef __CSOCP_KKT_H__
#define __CSOCP_KKT_H__

#include "csocp_config.h"
#ifdef CSOCP_WITH_DEBUG
#include <stdio.h>
#endif

/* KKT status codes */
#define KKT_OK 0
#define KKT_INVALID 1

/**
 * Creating and deleting methods
 */

/* Creating KKT structure */
void c_kkt_create(c_kkt **p_kkt, /* Pointer to KKT structure */
                  c_int nK,      /* Number of columns in KKT matrix */
#ifndef CSOCP_WITH_DEBUG
                  c_int nnzK /* Number of elements in KKT matrix */
#else
                  c_int nnzK,   /* Number of elements in KKT matrix */
                  FILE *dbgfile /* Debug log file */
#endif
);

/* Free KKT structure */
void c_kkt_free(c_kkt **p_kkt /* Pointer to KKT structure */
);

/**
 * Solve related methods
 */

/* Initializes the KKT system */
void c_kkt_init(c_kkt *kkt,    /* Input KKT structure */
                c_int n,       /* Number of variables */
                c_int p,       /* Number of inequalities */
                c_cone *cone,  /* Input cone */
                c_real *c,     /* Objective costs */
                c_real *b,     /* RHS of equalities */
                c_real *h,     /* RHS of inequalities */
                c_param *param /* Solving parametrs */
);

/* Symbolic factorization of KKT system */
c_int c_kkt_symbolic(c_kkt *kkt,       /* Input KKT structure */
                     c_spmat *ku,      /* Input KKT matrix */
                     c_int *sign,      /* Ouput sign vector */
                     c_buffer *buffer, /* Input buffer */
                     c_attr *attr      /* Input/Output attribute structure */
);

/* Numerical factorization of KKT system */
void c_kkt_factor(c_kkt *kkt,       /* Input KKT structure */
                  c_buffer *buffer, /* Input buffer */
                  c_param *param,   /* Input parameters */
                  c_attr *attr      /* Input attributes */
);

/* Solve the KKT system */
void c_kkt_solve(c_kkt *kkt,       /* Input KKT structure */
                 c_int n,          /* Number of variables */
                 c_int p,          /* Number of inequalities */
                 c_int m,          /* Number of equalities */
                 c_cone *cone,     /* Input cone structure */
                 c_spmat *A,       /* Input equality matrix */
                 c_spmat *G,       /* Input inequality matrix */
                 c_real *rhs,      /* RHS values */
                 c_real *dx,       /* Search direction of x-part */
                 c_real *dy,       /* Search direction of y-part */
                 c_real *dz,       /* Search direction of z-part */
                 c_int isinit,     /* Is initial KKT solve? */
                 c_buffer *buffer, /* Input buffer */
                 c_param *param,   /* Input parameters */
                 c_attr *attr      /* Input attributes */
);

/* Update the KKT system */
void c_kkt_update(c_kkt *kkt,    /* Input KKT structure */
                  c_cone *cone,  /* Input cone structure */
                  c_param *param /* Solving parameters */
);

/**
 * Auxiliary methods
 */

/* Get memory requirement of KKT */
c_int c_kkt_getmem(c_int n,    /* Number of variables */
                   c_int m,    /* Number of inequalities */
                   c_int p,    /* Number of equalities */
                   c_int nnzA, /* Number of nonzeros of A */
                   c_int nnzG, /* Number of nonzeros of G */
                   c_int nK,   /* Number of columns in KKT */
                   c_int nnzK  /* Number of elements in KKT */
);

#endif
