/**
 * This file implements Sparse cholesky factorization module in CSOCP solver
 */

#include "csocp_ldl.h"
#include "csocp_config.h"

#include "tool_math.h"
#include "tool_print.h"
#include "tool_time.h"

/* Symbolic factorization */
void c_ldl_symbolic(c_int ndim,   /* Dimension of KKT matrix */
                    c_int *Ajc,   /* Input begin pointer of KKT matrix */
                    c_int *Air,   /* Input index pointer of KKT matrix */
                    c_int *Ljc,   /* Output begin pointer of matrix */
                    c_int *etree, /* Output elimination tree */
                    c_int *lnz,   /* Temporary non-zeros count */
#ifndef CSOCP_WITH_DEBUG
                    c_int *flag /* Temporary visit flag */
#else
                    c_int *flag,  /* Temporary visit flag */
                    FILE *dbgfile /* Debug log file */
#endif
) {
  c_int i = 0, k = 0, p = 0;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(dbgfile, "\n");
  c_fprint(dbgfile, "[%s:%d]: Starting symbolic LDL factorization\n", __FILE__,
           __LINE__);
#endif

  for (k = 0; k < ndim; k++) {
    etree[k] = -1;
    flag[k] = k;
    lnz[k] = 0;

    for (p = Ajc[k]; p < Ajc[k + 1]; p++) {
      i = Air[p];

      for (; flag[i] != k; i = etree[i]) {
        if (etree[i] == -1) {
          etree[i] = k;
        }
        lnz[i]++;
        flag[i] = k;
      }
    }
  }

  Ljc[0] = 0;
  for (k = 0; k < ndim; k++) {
    Ljc[k + 1] = Ljc[k] + lnz[k];
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(dbgfile, "[%s:%d]: Finish symbolic LDL factorization (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(dbgfile, "\n");
#endif

  return;
}

/* Numerical factorization */
void c_ldl_numeric(c_int ndim,     /* Dimension of KKT matrix */
                   c_int *Ajc,     /* Input begin pointer of KKT matrix */
                   c_int *Air,     /* Input index pointer of KKT matrix */
                   c_real *Apr,    /* Input element pointer of KKT matrix */
                   c_int *Ljc,     /* Input begin pointer of L matrix */
                   c_int *etree,   /* Input elimination tree */
                   c_int *sign,    /* Input sign vector */
                   c_real eps,     /* Input regularization threshold */
                   c_real delta,   /* Input regularization parameter */
                   c_int *Lir,     /* Output index pointer of L matrix */
                   c_real *Lpr,    /* Output element pointer of L matrix */
                   c_real *D,      /* Output diagonal D matrix */
                   c_real *col,    /* Temporary column vector */
                   c_int *lnz,     /* Temporary non-zeros count of L matrix */
                   c_int *pattern, /* Temporary non-zeros pattern */
#ifndef CSOCP_WITH_DEBUG
                   c_int *flag /* Temporary flag for visit status */
#else
                   c_int *flag,   /* Temporary flag for visit status */
                   FILE *dbgfile  /* Debug log file */
#endif
) {
  c_int i = 0, k = 0, p = 0, l = 0, t = 0;
  c_real ci = 0.0, li = 0.0;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(dbgfile, "\n");
  c_fprint(dbgfile, "[%s:%d]: Starting numeric LDL factorization\n", __FILE__,
           __LINE__);
#endif

  for (k = 0; k < ndim; k++) {
    t = ndim;

    col[k] = 0.0;
    flag[k] = k;
    lnz[k] = 0;

    for (p = Ajc[k]; p < Ajc[k + 1]; p++) {
      i = Air[p];
      col[i] = Apr[p];

      for (l = 0; flag[i] != k; i = etree[i]) {
        pattern[l++] = i;
        flag[i] = k;
      }

      while (l > 0) {
        pattern[--t] = pattern[--l];
      }
    }

    D[k] = col[k];
    col[k] = 0.0;

    for (; t < ndim; t++) {
      i = pattern[t];
      ci = col[i];
      col[i] = 0.0;

      for (p = Ljc[i]; p < Ljc[i] + lnz[i]; p++) {
        col[Lir[p]] -= Lpr[p] * ci;
      }

#ifdef CSOCP_WITH_C6678
      li = c_math_divd(ci, D[i]);
#else
      li = ci / D[i];
#endif
      D[k] -= li * ci;

      Lir[p] = k;
      Lpr[p] = li;
      lnz[i]++;
    }

    D[k] = sign[k] * D[k] <= eps ? sign[k] * delta : D[k];
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(dbgfile, "[%s:%d]: Finish numeric LDL factorization (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(dbgfile, "\n");
#endif

  return;
}

/* Solve L*x = b */
void c_ldl_lsolve(c_int ndim,  /* Dimension of L matrix */
                  c_real *b,   /* Input RHS value */
                  c_int *Ljc,  /* Input begin pointer of L matrix */
                  c_int *Lir,  /* Inut index pointer of L matrix */
                  c_real *Lpr, /* Input element pointer of L matrix */
#ifndef CSOCP_WITH_DEBUG
                  c_real *x /* Output solution value */
#else
                  c_real *x,    /* Output solution value */
                  FILE *dbgfile /* Debug log file */
#endif
) {
  c_int i = 0, j = 0;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(dbgfile, "\n");
  c_fprint(dbgfile, "[%s:%d]: Starting forward solve L*x = b\n", __FILE__, __LINE__);
#endif

  for (i = 0; i < ndim; i++) {
    x[i] = b[i];
  }
  for (i = 0; i < ndim; i++) {
    for (j = Ljc[i]; j < Ljc[i + 1]; j++) {
      x[Lir[j]] -= Lpr[j] * x[i];
    }
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(dbgfile, "[%s:%d]: Finish forward solve L*x = b (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(dbgfile, "\n");
#endif

  return;
}

/* Solve D*x = b */
void c_ldl_dsolve(c_int ndim, /* Dimension of diagonal matrix */
                  c_real *x,  /* Input + Output value */
#ifndef CSOCP_WITH_DEBUG
                  c_real *D /* Input diagonal matrix */
#else
                  c_real *D,    /* Input diagonal matrix */
                  FILE *dbgfile /* Debug log file */
#endif
) {
  c_int i = 0;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(dbgfile, "\n");
  c_fprint(dbgfile, "[%s:%d]: Starting diagonal solve D*x = b\n", __FILE__, __LINE__);
#endif

  for (i = 0; i < ndim; i++) {
#ifdef CSOCP_WITH_C6678
    x[i] = c_math_divd(x[i], D[i]);
#else
    x[i] /= D[i];
#endif
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(dbgfile, "[%s:%d]: Finish diagonal solve D*x = b (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(dbgfile, "\n");
#endif

  return;
}

/* Solve L'*x = b */
void c_ldl_ltsolve(c_int ndim, /* Dimension of L matrix */
                   c_real *x,  /* Input + Output value */
                   c_int *Ljc, /* Input begin pointer of L matrix */
                   c_int *Lir, /* Inut index pointer of L matrix */
#ifndef CSOCP_WITH_DEBUG
                   c_real *Lpr /* Input element pointer of L matrix */
#else
                   c_real *Lpr,  /* Input element pointer of L matrix */
                   FILE *dbgfile /* Debug log file */
#endif
) {
  c_int i = 0, j = 0;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(dbgfile, "\n");
  c_fprint(dbgfile, "[%s:%d]: Starting backward solve L'*x = b\n", __FILE__, __LINE__);
#endif

  for (i = ndim - 1; i >= 0; i--) {
    for (j = Ljc[i]; j < Ljc[i + 1]; j++) {
      x[i] -= Lpr[j] * x[Lir[j]];
    }
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(dbgfile, "[%s:%d]: Finish backward solve L'*x = b (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(dbgfile, "\n");
#endif

  return;
}
