/**
 * This file implements parameter module for the CSOCP solver
 *
 * We use macros for default values of parameters, and also provide
 * methods for reseting and setting new values.
 */

#include "csocp.h"

#include "csocp_config.h"
#include "csocp_param.h"

#include "struct_param.h"

#ifdef CSOCP_WITH_ECOS
/* Default parameters */
#define MAXIT 200 /* Maximum number of iterations */

#define RLXFACTOR 500    /* Maximum factor for relax tolerance */
#define FEASTOL 1E-7     /* Primal/dual infeasibility tolerance */
#define ABSTOL 1E-7      /* Absolute tolerance on duality gap */
#define RELTOL 1E-7      /* Relative tolerance on duality gap */
#define FEASTOL_RLX 1E-4 /* Relaxed primal/dual feasibility tolerance */
#define ABSTOL_RLX 5E-5  /* Relaxed absolute tolerance */
#define RELTOL_RLX 5E-5  /* Relaxed relative tolerance */

#define SCALEMETHOD 0 /* Scaling method (-1: Auto, 0: Ruiz, 1: Pock-Chambolle) */
#define SCALEITER 3   /* Number of ruiz scaling iterations */
#define SCALETOL 1E-6 /* Scaling tolerance */

#define MUSTRATEGY 0  /* Mu update strategy */
#define ALPHALOQO 0.8 /* Alpha of LOQO strategy */

#define IRITER 9        /* Number of IR steps */
#define IRFACT 6        /* Factor of IR error */
#define LINSYSTOL 1E-14 /* Linear system tolerance */
#define DELTASTAT 7E-8  /* Static regularization parameter */
#define DELTA 2E-7      /* Dynamic regularization parameter */
#define EPSILON 1E-13   /* Epsilon tolerance */

#define GAMMA 0.99    /* Scaling factor of step length */
#define STEPMIN 1E-6  /* Smallest step allowed */
#define STEPMAX 0.999 /* Largest step allowed */
#define SIGMAMIN 1E-4 /* Always do some centering */
#define SIGMAMAX 1.0  /* Never fully center */
#else
/* Default parameters */
#define MAXIT 200 /* Maximum number of iterations */

#define RLXFACTOR 500    /* Maximum factor for relax tolerance */
#define FEASTOL 1E-7     /* Primal/dual infeasibility tolerance */
#define ABSTOL 1E-7      /* Absolute tolerance on duality gap */
#define RELTOL 1E-7      /* Relative tolerance on duality gap */
#define FEASTOL_RLX 1E-4 /* Relaxed primal/dual feasibility tolerance */
#define ABSTOL_RLX 5E-5  /* Relaxed absolute tolerance */
#define RELTOL_RLX 5E-5  /* Relaxed relative tolerance */

#define SCALEMETHOD 1 /* Scaling method (-1: Auto, 0: Ruiz, 1: Pock-Chambolle) */
#define SCALEITER 3   /* Number of ruiz scaling iterations */
#define SCALETOL 1E-6 /* Scaling tolerance */

#define MUSTRATEGY 0  /* Mu update strategy */
#define ALPHALOQO 0.8 /* Alpha of LOQO strategy */

#define IRITER 6        /* Number of IR steps */
#define IRFACT 3        /* Factor of IR error */
#define LINSYSTOL 1E-13 /* Linear system tolerance */
#define DELTASTAT 5E-8  /* Static regularization parameter */
#define DELTA 1E-7      /* Dynamic regularization parameter */
#define EPSILON 1E-13   /* Epsilon tolerance */

#define GAMMA 0.99    /* Scaling factor of step length */
#define STEPMIN 1E-6  /* Smallest step allowed */
#define STEPMAX 0.999 /* Largest step allowed */
#define SIGMAMIN 1E-4 /* Always do some centering */
#define SIGMAMAX 1.0  /* Never fully center */
#endif

/* Reset parameters to default values */
void c_param_reset(c_param *param /* Input parameter */
) {
  param->maxit = MAXIT;

  param->rlxfactor = RLXFACTOR;
  param->feastol = FEASTOL;
  param->abstol = ABSTOL;
  param->reltol = RELTOL;
  param->feastol_rlx = FEASTOL_RLX;
  param->abstol_rlx = ABSTOL_RLX;
  param->reltol_rlx = RELTOL_RLX;

  param->scalemethod = SCALEMETHOD;
  param->scaleiter = SCALEITER;
  param->scaletol = SCALETOL;

  param->mustrategy = MUSTRATEGY;
  param->alphaloqo = ALPHALOQO;

  param->iriter = IRITER;
  param->irfact = IRFACT;
  param->linsystol = LINSYSTOL;
  param->deltastat = DELTASTAT;
  param->delta = DELTA;
  param->epsilon = EPSILON;

  param->gamma = GAMMA;
  param->minstep = STEPMIN;
  param->maxstep = STEPMAX;
  param->minsigma = SIGMAMIN;
  param->maxsigma = SIGMAMAX;
}

/* Set double parameter */
void c_param_setdbl(c_param *param, /* Input parameter */
                    c_int dblparam, /* ID of double parameter */
                    c_real newval   /* New value */
) {
  switch (dblparam) {
  case CSOCP_DBLPARAM_FEASTOL:
    param->feastol = newval;
    break;
  case CSOCP_DBLPARAM_ABSTOL:
    param->abstol = newval;
    break;
  case CSOCP_DBLPARAM_RELTOL:
    param->reltol = newval;
    break;
  case CSOCP_DBLPARAM_FEASTOLRLX:
    param->feastol_rlx = newval;
    break;
  case CSOCP_DBLPARAM_ABSTOLRLX:
    param->abstol_rlx = newval;
    break;
  case CSOCP_DBLPARAM_RELTOLRLX:
    param->reltol_rlx = newval;
    break;
  case CSOCP_DBLPARAM_SCALETOL:
    param->scaletol = newval;
    break;
  default:
    break;
  }

  return;
}

/* Set integer parameter */
void c_param_setint(c_param *param, /* Input parameter */
                    c_int intparam, /* ID of integer parameter */
                    c_int newval    /* New value */
) {
  switch (intparam) {
  case CSOCP_INTPARAM_ITERLIMIT:
    param->maxit = newval;
    break;
  case CSOCP_INTPARAM_SCALEMETHOD:
    if (newval < 0 || newval > 1) {
      param->scalemethod = 0;
    } else {
      param->scalemethod = newval;
    }
    break;
  case CSOCP_INTPARAM_SCALEITER:
    param->scaleiter = newval;
    break;
  case CSOCP_INTPARAM_MUSTRATEGY:
    if (newval < 0 || newval > 1) {
      param->mustrategy = 0;
    } else {
      param->mustrategy = newval;
    }
    break;
  default:
    break;
  }

  return;
}

/* Get double parameter */
c_real c_param_getdbl(c_param *param, /* Input parameter */
                      c_int dblparam  /* ID of double parameter */
) {
  c_real retval = CSOCP_INFINITY;

  switch (dblparam) {
  case CSOCP_DBLPARAM_FEASTOL:
    retval = param->feastol;
    break;
  case CSOCP_DBLPARAM_ABSTOL:
    retval = param->abstol;
    break;
  case CSOCP_DBLPARAM_RELTOL:
    retval = param->reltol;
    break;
  case CSOCP_DBLPARAM_FEASTOLRLX:
    retval = param->feastol_rlx;
    break;
  case CSOCP_DBLPARAM_ABSTOLRLX:
    retval = param->abstol_rlx;
    break;
  case CSOCP_DBLPARAM_RELTOLRLX:
    retval = param->reltol_rlx;
    break;
  case CSOCP_DBLPARAM_SCALETOL:
    retval = param->scaletol;
    break;
  default:
    break;
  }

  return retval;
}

/* Get integer parameter */
c_int c_param_getint(c_param *param, /* Input parameter */
                     c_int intparam  /* ID of integer parameter */
) {
  c_int retval = -1;

  switch (intparam) {
  case CSOCP_INTPARAM_ITERLIMIT:
    retval = param->maxit;
    break;
  case CSOCP_INTPARAM_SCALEMETHOD:
    retval = param->scalemethod;
    break;
  case CSOCP_INTPARAM_SCALEITER:
    retval = param->scaleiter;
    break;
  case CSOCP_INTPARAM_MUSTRATEGY:
    retval = param->mustrategy;
    break;
  default:
    break;
  }

  return retval;
}

/* Get memory requirement of parameter */
c_int c_param_getmem() { return sizeof(c_param); }
