/**
 * This file implements routines for scaling methods in CSOCP solver
 *
 * This module serves as part of presolving module
 */

#include "csocp.h"

#include "csocp_config.h"
#include "csocp_scaling.h"

#include "tool_buffer.h"
#include "tool_math.h"
#include "tool_memory.h"
#include "tool_print.h"
#include "tool_spmat.h"
#include "tool_time.h"

#include "struct_attr.h"
#include "struct_cone.h"
#include "struct_param.h"
#include "struct_prob.h"
#include "struct_spmat.h"

#include <stdio.h>
#include <stdlib.h>

/**
 * Scaling methods
 */

/* Compute scaling of SOC cones */
static void scaling_soc(c_cone *cone, c_real *srcDstVec) {
  c_int i = 0, iconeidx = 0;
  c_real dnorm = 0.0;

  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

  iconeidx = lp->ndim;
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    /* TODO: Better scaling of SOC cones */
    dnorm = c_math_sum(soc->ndim, srcDstVec + iconeidx);
    c_math_filldbl(soc->ndim, dnorm, srcDstVec + iconeidx);

    iconeidx += soc->ndim;
  }

  return;
}

/* Scale the problem with ruiz method */
static void c_scaling_ruiz(csocp_prob *prob) {
  c_int i = 0, iter = 0;

  c_spmat *matA = prob->A;
  c_spmat *matG = prob->G;

  c_param *param = prob->param;
  c_buffer *buffer = prob->buffer;

  c_int nCol = matA != NULL ? matA->nCol : matG->nCol;
  c_int nRowA = matA != NULL ? matA->nRow : 0;
  c_int nRowG = matG->nRow;

  c_real *tmpX = NULL;
  c_real *tmpA = NULL;
  c_real *tmpG = NULL;

#ifdef CSOCP_WITH_DEBUG
  /* Start timer for ruiz scaling */
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting ruiz scaling\n", __FILE__, __LINE__);
#endif

  /* Obtain temporary memory for buffer */
  c_buffer_request_dbuf(buffer, tmpX, nCol);
  c_buffer_request_dbuf(buffer, tmpA, nRowA);
  c_buffer_request_dbuf(buffer, tmpG, nRowG);

  /* Main loop of scaling iteration */
  for (iter = 0; iter < param->scaleiter; ++iter) {
    /* Initialize scaling factors to zeros */
    c_math_filldbl(nCol, 0.0, tmpX);
    c_math_filldbl(nRowA, 0.0, tmpA);
    c_math_filldbl(nRowG, 0.0, tmpG);

    /* Compute inf-norm of columns */
    if (matA != NULL) {
      c_spmat_infnorm(matA, tmpX, 1);
    }
    if (nRowG > 0) {
      c_spmat_infnorm(matG, tmpX, 1);
    }

    /* Get the norms */
    for (i = 0; i < nCol; ++i) {
      tmpX[i] = tmpX[i] < param->scaletol ? 1.0 : c_math_sqrt(tmpX[i]);
    }

    /* Compute inf-norm of rows */
    if (matA != NULL) {
      c_spmat_infnorm(matA, tmpA, 0);

      /* Get the norms */
      for (i = 0; i < nRowA; i++) {
        tmpA[i] = tmpA[i] < param->scaletol ? 1.0 : c_math_sqrt(tmpA[i]);
      }
    }

    if (nRowG > 0) {
      c_spmat_infnorm(matG, tmpG, 0);

      /* TODO: Scaling cone variables */
      scaling_soc(prob->cone, tmpG);

      /* Get the norms */
      for (i = 0; i < nRowG; i++) {
        tmpG[i] = tmpG[i] < param->scaletol ? 1.0 : c_math_sqrt(tmpG[i]);
      }
    }

    /* Scale rows of matrix */
    if (matA != NULL) {
      c_spmat_setscaling(matA, tmpA, 0);
    }
    if (nRowG > 0) {
      c_spmat_setscaling(matG, tmpG, 0);
    }

    /* Scale columns of matrix */
    if (matA != NULL) {
      c_spmat_setscaling(matA, tmpX, 1);
    }
    if (nRowG > 0) {
      c_spmat_setscaling(matG, tmpX, 1);
    }

    /* Update scaling factors */
    c_math_mul(nCol, tmpX, prob->facX);
    c_math_mul(nRowA, tmpA, prob->facA);
    c_math_mul(nRowG, tmpG, prob->facG);
  }

  /* Release temporary buffer */
  c_buffer_release_dbuf(buffer, nCol);  /* tmpX */
  c_buffer_release_dbuf(buffer, nRowA); /* tmpA */
  c_buffer_release_dbuf(buffer, nRowG); /* tmpG */

#ifdef CSOCP_WITH_DEBUG
  /* Compute elapsed time for ruiz scaling */
  c_fprint(prob->dbgfile, "[%s:%d]: Finish ruiz scaling (%.4fs)\n", __FILE__, __LINE__,
           c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif

#ifdef CSOCP_WITH_DUMP
  /* Dump scaling factors to file */
  c_math_dumpdbl(nCol, prob->facX, "facX_ruiz.txt");
  c_math_dumpdbl(nRowA, prob->facA, "facA_ruiz.txt");
  c_math_dumpdbl(nRowG, prob->facG, "facG_ruiz.txt");
#endif

  return;
}

/* Scale the problem with pock-chambolle method */
static void c_scaling_pock(csocp_prob *prob) {
  c_int i = 0;

  c_spmat *matA = prob->A;
  c_spmat *matG = prob->G;

  c_param *param = prob->param;
  c_buffer *buffer = prob->buffer;

  c_int nCol = matA != NULL ? matA->nCol : matG->nCol;
  c_int nRowA = matA != NULL ? matA->nRow : 0;
  c_int nRowG = matG->nRow;

  c_real *tmpX = NULL;
  c_real *tmpA = NULL;
  c_real *tmpG = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting pock-ruiz scaling\n", __FILE__, __LINE__);
#endif

  /* Obtain temporary memory for buffer */
  c_buffer_request_dbuf(buffer, tmpX, nCol);
  c_buffer_request_dbuf(buffer, tmpA, nRowA);
  c_buffer_request_dbuf(buffer, tmpG, nRowG);

  c_math_filldbl(nCol, 0.0, tmpX);
  c_math_filldbl(nRowA, 0.0, tmpA);
  c_math_filldbl(nRowG, 0.0, tmpG);

  /* Compute 1-norm of columns */
  if (matA != NULL) {
    c_spmat_onenorm(matA, tmpX, 1);
  }
  if (nRowG > 0) {
    c_spmat_onenorm(matG, tmpX, 1);
  }

  /* Get the norms */
  for (i = 0; i < nCol; ++i) {
    tmpX[i] = tmpX[i] < param->scaletol ? 1.0 : c_math_sqrt(tmpX[i]);
  }

  /* Compute 1-norm of rows */
  if (matA != NULL) {
    c_spmat_onenorm(matA, tmpA, 0);

    /* Get the norms */
    for (i = 0; i < nRowA; ++i) {
      tmpA[i] = tmpA[i] < param->scaletol ? 1.0 : c_math_sqrt(tmpA[i]);
    }
  }

  if (nRowG > 0) {
    c_spmat_onenorm(matG, tmpG, 0);

    /* TODO: Scaling cone variables */
    scaling_soc(prob->cone, tmpG);

    /* Get the norms */
    for (i = 0; i < nRowG; i++) {
      tmpG[i] = tmpG[i] < param->scaletol ? 1.0 : c_math_sqrt(tmpG[i]);
    }
  }

  /* Scale rows of matrix */
  if (matA != NULL) {
    c_spmat_setscaling(matA, tmpA, 0);
  }
  if (nRowG > 0) {
    c_spmat_setscaling(matG, tmpG, 0);
  }

  /* Scale columns of matrix */
  if (matA != NULL) {
    c_spmat_setscaling(matA, tmpX, 1);
  }
  if (nRowG > 0) {
    c_spmat_setscaling(matG, tmpX, 1);
  }

  /* Update scaling factors */
  c_math_mul(nCol, tmpX, prob->facX);
  c_math_mul(nRowA, tmpA, prob->facA);
  c_math_mul(nRowG, tmpG, prob->facG);

  /* Release temporary buffer */
  c_buffer_release_dbuf(buffer, nCol);  /* tmpX */
  c_buffer_release_dbuf(buffer, nRowA); /* tmpA */
  c_buffer_release_dbuf(buffer, nRowG); /* tmpG */

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish pock-ruiz scaling (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif

#ifdef CSOCP_WITH_DUMP
  /* Dump scaling factors to file */
  c_math_dumpdbl(nCol, prob->facX, "facX_pock.txt");
  c_math_dumpdbl(nRowA, prob->facA, "facA_pock.txt");
  c_math_dumpdbl(nRowG, prob->facG, "facG_pock.txt");
#endif

  return;
}

/* Scale the problem */
void c_scaling_set(csocp_prob *prob) {
#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting problem scaling\n", __FILE__, __LINE__);
#endif

  /* Initialize scaling factors to 1.0 */
  c_math_filldbl(prob->n, 1.0, prob->facX);
  c_math_filldbl(prob->p, 1.0, prob->facA);
  c_math_filldbl(prob->m, 1.0, prob->facG);

  if (prob->param->scalemethod == -1) {
    if (prob->attr->nnzK > (1 << 16)) {
      /* Use ruiz scaling */
      c_scaling_ruiz(prob);
    } else {
      /* Use pock-chambolle scaling */
      c_scaling_pock(prob);
      c_scaling_ruiz(prob);
    }
  } else if (prob->param->scalemethod == 0) {
    /* Use ruiz scaling */
    c_scaling_ruiz(prob);
  } else {
    /* Use pock-chambolle scaling */
    c_scaling_pock(prob);
    c_scaling_ruiz(prob);
  }

  /* Scaling the right-hand-side */
  c_math_div(prob->n, prob->facX, prob->c);
  c_math_div(prob->p, prob->facA, prob->b);
  c_math_div(prob->m, prob->facG, prob->h);

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish problem scaling (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif
}

/* Unscale the problem */
void c_scaling_unset(csocp_prob *prob /* The CSOCP problem */
) {
#ifndef CSOCP_WITH_COPTPROB
#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting problem unscaling\n", __FILE__, __LINE__);
#endif

  c_math_mul(prob->n, prob->facX, prob->c);
  c_math_mul(prob->p, prob->facA, prob->b);
  c_math_mul(prob->m, prob->facG, prob->h);

  if (prob->A != NULL) {
    c_spmat_unsetscaling(prob->A, prob->facA, prob->facX);
  }
  if (prob->G->nRow > 0) {
    c_spmat_unsetscaling(prob->G, prob->facG, prob->facX);
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish problem unscaling (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif
#endif
  return;
}