/**
 * This file implements solver module of the CSOCP solver
 */

#include "csocp_solver.h"
#include "csocp_attr.h"
#include "csocp_cone.h"
#include "csocp_config.h"
#include "csocp_kkt.h"
#include "csocp_param.h"

#include "struct_attr.h"
#include "struct_buffer.h"
#include "struct_cone.h"
#include "struct_kkt.h"
#include "struct_param.h"
#include "struct_prob.h"
#include "struct_spmat.h"

#include "tool_buffer.h"
#include "tool_math.h"
#include "tool_print.h"
#include "tool_spmat.h"
#include "tool_time.h"

#include <math.h>
#include <stdlib.h>

/* Compute the starting point */
c_int c_solver_startpoint(csocp_prob *prob) {
  c_kkt *kkt = prob->kkt;

  c_param *param = prob->param;
  c_attr *attr = prob->attr;

  c_buffer *buffer = prob->buffer;

  c_real *work_r = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting computing initial point\n", __FILE__,
           __LINE__);
#endif

  /* Compute initial residuals */
  prob->resy0 = CSOCP_MAX(1.0, c_math_norm2(prob->p, prob->b));
  prob->resx0 = CSOCP_MAX(1.0, c_math_norm2(prob->n, prob->c));
  prob->resz0 = CSOCP_MAX(1.0, c_math_norm2(prob->m, prob->h));

  /* Initialize kappa and tau */
  attr->kap = 1.0;
  attr->tau = 1.0;

  /* Initialize the KKT system */
  c_kkt_init(kkt, prob->n, prob->p, prob->cone, prob->c, prob->b, prob->h, param);

  /* Factor KKT matrix */
  c_kkt_factor(kkt, buffer, param, attr);

  /* First pass: Compute primal start values [xhat; y; -r] with [0; b; h] */
  c_kkt_solve(kkt, prob->n, prob->p, prob->m, prob->cone, prob->A, prob->G, kkt->rhs1,
              prob->dx1, prob->dy1, prob->dz1, 1, buffer, param, attr);
#ifdef CSOCP_WITH_DUMP
  /* Dump search direction */
  c_math_dumpdbl(prob->n, prob->dx1, "dx1_init.txt");
  c_math_dumpdbl(prob->p, prob->dy1, "dy1_init.txt");
  c_math_dumpdbl(prob->m, prob->dz1, "dz1_init.txt");
#endif

  /* Copy initial value of x */
  c_math_copydbl(prob->n, prob->dx1, prob->x, 0);

  /* Get temporary memory from buffer */
  c_buffer_request_dbuf(buffer, work_r, attr->nK);

  /* Copy -r */
  c_math_copydbl(prob->m, prob->dz1, work_r, 1);

  /* Compute s, make sure they are inside cones */
  c_cone_pushbnd(prob->cone, work_r, prob->s, param);

  /* Release buffer */
  c_buffer_release_dbuf(buffer, attr->nK); /* work_r */

  /* Second pass: Compute dual start values [x; yhat; zbar] with [-c; 0; 0] */
  c_kkt_solve(kkt, prob->n, prob->p, prob->m, prob->cone, prob->A, prob->G, kkt->rhs2,
              prob->dx2, prob->dy2, prob->dz2, 1, buffer, param, attr);
#ifdef CSOCP_WITH_DUMP
  /* Dump search direction */
  c_math_dumpdbl(prob->n, prob->dx2, "dx2_init.txt");
  c_math_dumpdbl(prob->p, prob->dy2, "dy2_init.txt");
  c_math_dumpdbl(prob->m, prob->dz2, "dz2_init.txt");
#endif

  /* Copy y */
  c_math_copydbl(prob->p, prob->dy2, prob->y, 0);

  /* Compute z */
  c_cone_pushbnd(prob->cone, prob->dz2, prob->z, param);

  /* Build new RHS1: [-c; b; h] */
  c_math_copydblbyidx(prob->n, kkt->invmap, prob->c, kkt->rhs1, 1);

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish computing initial point (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif

  return 0;
}

/* Line-search */
static c_real solver_linesearch(csocp_prob *prob, c_real dtau, c_real dkap) {
  c_int i = 0, j = 0;
  c_int iconeidx = 0;

  c_real alpha = 0.0, step = 0.0;
  c_real rhomin = 0.0, sigmamin = 0.0;
  c_real rhonorm = 0.0, sigmanorm = 0.0, lknorm = 0.0;
  c_real rhofactor = 0.0, sigmfactor = 0.0;
  c_real lkbar_ds = 0.0, lkbar_dz = 0.0;

  c_real *lk = NULL, *dsk = NULL, *dzk = NULL;

  c_kkt *kkt = prob->kkt;

  c_param *param = prob->param;
  c_attr *attr = prob->attr;

  c_buffer *buffer = prob->buffer;

  c_cone *cone = prob->cone;
  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

  c_real *lambda = prob->lambda;
  c_real *ds = prob->dsaff_w;
  c_real *dz = prob->dzaff_w;

#ifdef CSOCP_WITH_C6678
  c_real tau_ratio = -c_math_divd(attr->tau, dtau);
  c_real kap_ratio = -c_math_divd(attr->kap, dkap);
#else
  c_real tau_ratio = -attr->tau / dtau;
  c_real kap_ratio = -attr->kap / dkap;
#endif

  c_real *lkbar = NULL;
  c_real *rho = NULL;
  c_real *sigma = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting line search\n", __FILE__, __LINE__);
#endif

  /* Get temporary memory from buffer */
  c_buffer_request_dbuf(buffer, lkbar, attr->nK);
  c_buffer_request_dbuf(buffer, rho, attr->nK);
  c_buffer_request_dbuf(buffer, sigma, attr->nK);

  /* Search on LP cone */
  if (lp->ndim > 0) {
#ifdef CSOCP_WITH_C6678
    rhomin = c_math_divd(ds[0], lambda[0]);
    sigmamin = c_math_divd(dz[0], lambda[0]);
    for (i = 1; i < lp->ndim; i++) {
      rhomin = CSOCP_MIN(c_math_divd(ds[i], lambda[i]), rhomin);
      sigmamin = CSOCP_MIN(c_math_divd(dz[i], lambda[i]), sigmamin);
    }

    if (-sigmamin > -rhomin) {
      alpha = sigmamin < 0 ? -c_math_divd(1.0, sigmamin) : 1e13;
    } else {
      alpha = rhomin < 0 ? -c_math_divd(1.0, rhomin) : 1e13;
    }
#else
    rhomin = ds[0] / lambda[0];
    sigmamin = dz[0] / lambda[0];
    for (i = 1; i < lp->ndim; i++) {
      rhomin = CSOCP_MIN(ds[i] / lambda[i], rhomin);
      sigmamin = CSOCP_MIN(dz[i] / lambda[i], sigmamin);
    }

    if (-sigmamin > -rhomin) {
      alpha = sigmamin < 0 ? -1.0 / sigmamin : 1e13;
    } else {
      alpha = rhomin < 0 ? -1.0 / rhomin : 1e13;
    }
#endif
  } else {
    alpha = 10.0;
  }

  /* Search on tau and kappa */
  if (tau_ratio > 0 && tau_ratio < alpha) {
    alpha = tau_ratio;
  }
  if (kap_ratio > 0 && kap_ratio < alpha) {
    alpha = kap_ratio;
  }

  /* Search on SOC cone */
  iconeidx = lp->ndim;
  for (i = 0; i < cone->nsoc; i++) {
    soc = cone->soc[i];

    lk = lambda + iconeidx;
    dsk = ds + iconeidx;
    dzk = dz + iconeidx;

    /* Compute norm */
    lknorm = lk[0] * lk[0] - c_math_dot(soc->ndim - 1, lk + 1, lk + 1);
    if (lknorm <= 0.0) {
      continue;
    }

    lknorm = c_math_sqrt(lknorm);
#ifdef CSOCP_WITH_C6678
    lknorm = c_math_divd(1.0, lknorm);
#else
    lknorm = 1.0 / lknorm;
#endif
    c_math_copydbl(soc->ndim, lk, lkbar, 0);
    c_math_scalemul(soc->ndim, lknorm, lkbar);

    /* Compute rho_k and sigma_k */
    lkbar_ds = lkbar[0] * dsk[0] - c_math_dot(soc->ndim - 1, lkbar + 1, dsk + 1);
    lkbar_dz = lkbar[0] * dzk[0] - c_math_dot(soc->ndim - 1, lkbar + 1, dzk + 1);

#ifdef CSOCP_WITH_C6678
    rhofactor = c_math_divd(lkbar_ds + dsk[0], lkbar[0] + 1);
    sigmfactor = c_math_divd(lkbar_dz + dzk[0], lkbar[0] + 1);
#else
    rhofactor = (lkbar_ds + dsk[0]) / (lkbar[0] + 1);
    sigmfactor = (lkbar_dz + dzk[0]) / (lkbar[0] + 1);
#endif

    rho[0] = lknorm * lkbar_ds;
    for (j = 1; j < soc->ndim; j++) {
      rho[j] = lknorm * (dsk[j] - rhofactor * lkbar[j]);
    }
    rhonorm = c_math_norm2(soc->ndim - 1, rho + 1) - rho[0];

    sigma[0] = lknorm * lkbar_dz;
    for (j = 1; j < soc->ndim; j++) {
      sigma[j] = lknorm * (dzk[j] - sigmfactor * lkbar[j]);
    }
    sigmanorm = c_math_norm2(soc->ndim - 1, sigma + 1) - sigma[0];

    /* Update alpha */
    step = 0;
    step = CSOCP_MAX(rhonorm, step);
    step = CSOCP_MAX(sigmanorm, step);
    if (step != 0) {
#ifdef CSOCP_WITH_C6678
      alpha = CSOCP_MIN(alpha, c_math_divd(1.0, step));
#else
      alpha = CSOCP_MIN(alpha, 1.0 / step);
#endif
    }

    iconeidx += soc->ndim;
  }

  /* Saturate between min/max step */
  alpha = CSOCP_MIN(alpha, param->maxstep);
  alpha = CSOCP_MAX(alpha, param->minstep);

  /* Release buffer */
  c_buffer_release_dbuf(buffer, attr->nK); /* lkbar */
  c_buffer_release_dbuf(buffer, attr->nK); /* rho */
  c_buffer_release_dbuf(buffer, attr->nK); /* sigma */

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish line search (%.4fs)\n", __FILE__, __LINE__,
           c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif

  return alpha;
}

/* Predicator step */
void c_solver_predictor(csocp_prob *prob, c_real *p_res) {
  c_int i = 0, idx = 0;
  c_real cx1 = 0.0, by1 = 0.0, hz1 = 0.0;
  c_real cx2 = 0.0, by2 = 0.0, hz2 = 0.0;
  c_real *sidx = NULL, *zidx = NULL;

  c_kkt *kkt = prob->kkt;
  c_int *pinv = kkt->invmap;
  c_real *rhs = kkt->rhs2;

  c_attr *attr = prob->attr;
  c_param *param = prob->param;

  c_buffer *buffer = prob->buffer;

  c_cone *cone = prob->cone;
  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting predictor step\n", __FILE__, __LINE__);
#endif

  /* Update RHS of KKT */
  c_math_copydblbyidx(prob->n, pinv, prob->rx, rhs, 0);
  pinv += prob->n;
  c_math_copydblbyidx(prob->p, pinv, prob->ry, rhs, 1);
  pinv += prob->p;

  c_math_copywithopidx(lp->ndim, pinv, prob->s, prob->rz, rhs, -1);
  pinv += lp->ndim;

  idx = lp->ndim;
  for (i = 0; i < cone->nsoc; ++i) {
    soc = cone->soc[i];

    sidx = prob->s + idx;
    zidx = prob->rz + idx;

    c_math_copywithopidx(soc->ndim, pinv, sidx, zidx, rhs, -1);
    pinv += soc->ndim;

    rhs[pinv[0]] = 0;
    rhs[pinv[1]] = 0;
    pinv += 2;

    idx += soc->ndim;
  }

  /* Update search direction */
  c_kkt_solve(kkt, prob->n, prob->p, prob->m, prob->cone, prob->A, prob->G, kkt->rhs2,
              prob->dx2, prob->dy2, prob->dz2, 0, buffer, param, attr);

  /* c'*x1 */
  cx1 = c_math_dot(prob->n, prob->c, prob->dx1);
  /* b'*y1 */
  by1 = c_math_dot(prob->p, prob->b, prob->dy1);
  /* h'*z1 */
  hz1 = c_math_dot(prob->m, prob->h, prob->dz1);

  *p_res = cx1 + by1 + hz1;

  /* c'*x2 */
  cx2 = c_math_dot(prob->n, prob->c, prob->dx2);
  /* b'*y2 */
  by2 = c_math_dot(prob->p, prob->b, prob->dy2);
  /* c'*z2 */
  hz2 = c_math_dot(prob->m, prob->h, prob->dz2);

  /* Update 'tauaff' */
  attr->tauaff = attr->cx + attr->by + attr->hz;
  attr->tauaff += cx2 + by2 + hz2;
#ifdef CSOCP_WITH_C6678
  attr->tauaff = c_math_divd(attr->tauaff, attr->kap_t - cx1 - by1 - hz1);
#else
  attr->tauaff /= (attr->kap_t - cx1 - by1 - hz1);
#endif

  /* Update 'kapaff' */
  attr->kapaff = -attr->kap - attr->kap_t * attr->tauaff;

  /* dz2 = dz2 + tauaff*dz1 */
  c_math_add(prob->m, attr->tauaff, prob->dz1, prob->dz2);
  /* dzaff_w = w*dz2 */
  c_cone_scaledual(prob->cone, prob->dz2, prob->dzaff_w);

  /* dsaff_w = -dzaff_w - lambda */
  c_math_copydbl(prob->m, prob->dzaff_w, prob->dsaff_w, 1);
  c_math_sub(prob->m, 1.0, prob->lambda, prob->dsaff_w);

  /* Perform line search on affine direction */
  attr->step_aff = solver_linesearch(prob, attr->tauaff, attr->kapaff);

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish predictor step (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif

  return;
}

/* Centering step */
void c_solver_centering(csocp_prob *prob) {
  c_attr *attr = prob->attr;
  c_param *param = prob->param;

  c_int i = 0;
  c_real cmpl = 0.0;
  c_real mincmpl = CSOCP_INFINITY;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting centering step\n", __FILE__, __LINE__);
#endif

  if (prob->param->mustrategy == 0) {
    attr->sigma = 1.0 - attr->step_aff;
    attr->sigma = attr->sigma * attr->sigma * attr->sigma;
    attr->sigma = CSOCP_MIN(attr->sigma, param->maxsigma);
    attr->sigma = CSOCP_MAX(attr->sigma, param->minsigma);
  } else if (prob->param->mustrategy == 1) {
    cmpl = c_math_dot(prob->m, prob->s, prob->z);
    for (i = 0; i < prob->m; ++i) {
      mincmpl = CSOCP_MIN(mincmpl, prob->s[i] * prob->z[i]);
    }
    mincmpl = CSOCP_MIN(mincmpl, attr->kapaff * attr->tauaff);
#ifdef CSOCP_WITH_C6678
    attr->sigma = c_math_divd(prob->m * mincmpl, cmpl + attr->kapaff * attr->tauaff);
#else
    attr->sigma = prob->m * mincmpl / (cmpl + attr->kapaff * attr->tauaff);
#endif
    attr->sigma = CSOCP_MIN(attr->sigma, param->maxsigma);
    attr->sigma = CSOCP_MAX(attr->sigma, param->minsigma);
  }

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish centering step (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif

  return;
}

/* Corrector step */
void c_solver_corrector(csocp_prob *prob, c_real res) {
  c_int i = 0, j = 0, idx = 0;
  c_real cx2 = 0.0, by2 = 0.0, hz2 = 0.0;
  c_real bkap = 0.0, rt = 0.0;
  c_real dtau = 0.0, dkap = 0.0;

  c_kkt *kkt = prob->kkt;
  c_int *pinv = kkt->invmap;
  c_real *rhs = kkt->rhs2;

  c_real *ds1 = NULL;
  c_real *ds2 = NULL;

  c_attr *attr = prob->attr;
  c_param *param = prob->param;

  c_buffer *buffer = prob->buffer;

  c_cone *cone = prob->cone;
  c_lp *lp = cone->lp;
  c_soc *soc = NULL;

  c_real sigmamu = attr->sigma * attr->mu;
  c_real sigmaone = 1.0 - attr->sigma;

#ifdef CSOCP_WITH_DEBUG
  c_real timebeg = c_timer_elapsed();
  c_fprint(prob->dbgfile, "\n");
  c_fprint(prob->dbgfile, "[%s:%d]: Starting corrector step\n", __FILE__, __LINE__);
#endif

  /* Obtain temporary memory from buffer */
  c_buffer_request_dbuf(buffer, ds1, attr->nK);
  c_buffer_request_dbuf(buffer, ds2, attr->nK);

  /* ds = lambda o lambda + dsaff_w o dzaff_w - sigma*mu */
  c_cone_prod(cone, prob->lambda, prob->lambda, ds1);
  c_cone_prod(cone, prob->dsaff_w, prob->dzaff_w, ds2);

  c_math_add(lp->ndim, 1.0, ds2, ds1);
  c_math_sub(lp->ndim, sigmamu, NULL, ds1);

  idx = lp->ndim;
  for (i = 0; i < cone->nsoc; i++) {
    soc = cone->soc[i];

    ds1[idx] += ds2[idx] - sigmamu;
    c_math_add(soc->ndim - 1, 1.0, ds2 + idx + 1, ds1 + idx + 1);

    idx += soc->ndim;
  }

  /* dz = -(1 - sigma)*rz + w * (lambda \ ds) */
  c_cone_div(cone, prob->lambda, ds1, prob->dsaff_w);
  c_cone_scaledual(cone, prob->dsaff_w, ds1);

  /* Update RHS */
  c_math_scalewithidx(prob->n, sigmaone, pinv, rhs);
  pinv += prob->n;
  c_math_scalewithidx(prob->p, sigmaone, pinv, rhs);
  pinv += prob->p;

  c_math_add(lp->ndim, -sigmaone, prob->rz, ds1);
  c_math_copydblbyidx(lp->ndim, pinv, ds1, rhs, 0);
  pinv += lp->ndim;

  idx = lp->ndim;
  for (j = 0; j < cone->nsoc; j++) {
    soc = cone->soc[j];

    c_math_add(soc->ndim, -sigmaone, prob->rz + idx, ds1 + idx);
    c_math_copydblbyidx(soc->ndim, pinv, ds1 + idx, rhs, 0);
    pinv += soc->ndim;

    rhs[pinv[0]] = 0;
    rhs[pinv[0]] = 0;
    pinv += 2;

    idx += soc->ndim;
  }

  /* Release buffer */
  c_buffer_release_dbuf(buffer, attr->nK); /* ds1 */
  c_buffer_release_dbuf(buffer, attr->nK); /* ds2 */

  c_kkt_solve(kkt, prob->n, prob->p, prob->m, prob->cone, prob->A, prob->G, kkt->rhs2,
              prob->dx2, prob->dy2, prob->dz2, 0, buffer, param, attr);

  cx2 = c_math_dot(prob->n, prob->c, prob->dx2);
  by2 = c_math_dot(prob->p, prob->b, prob->dy2);
  hz2 = c_math_dot(prob->m, prob->h, prob->dz2);

  /* Update 'dtau' */
  bkap = attr->kap * attr->tau + attr->kapaff * attr->tauaff - sigmamu;
  rt = attr->kap + attr->cx + attr->by + attr->hz;

#ifdef CSOCP_WITH_C6678
  dtau = sigmaone * rt - c_math_divd(bkap, attr->tau);
  dtau += cx2 + by2 + hz2;
  dtau = c_math_divd(dtau, attr->kap_t - res);
#else
  dtau = sigmaone * rt - bkap / attr->tau;
  dtau += cx2 + by2 + hz2;
  dtau /= (attr->kap_t - res);
#endif

  /* Update 'dkap' */
#ifdef CSOCP_WITH_C6678
  dkap = -c_math_divd(bkap + attr->kap * dtau, attr->tau);
#else
  dkap = -(bkap + attr->kap * dtau) / attr->tau;
#endif

  /* Update search direction */
  c_math_add(prob->n, dtau, prob->dx1, prob->dx2);
  c_math_add(prob->p, dtau, prob->dy1, prob->dy2);
  c_math_add(prob->m, dtau, prob->dz1, prob->dz2);

  c_cone_scaledual(prob->cone, prob->dz2, prob->dzaff_w);
  c_math_add(prob->m, 1.0, prob->dzaff_w, prob->dsaff_w);
  c_math_scalemul(prob->m, -1.0, prob->dsaff_w);

  /* Perform line search on combined direction */
  attr->step = solver_linesearch(prob, dtau, dkap) * param->gamma;

  /* Bring ds to unscaled form */
  c_cone_scaledual(prob->cone, prob->dsaff_w, prob->dsaff);

  /* Update solutions by search step */
  c_math_add(prob->n, attr->step, prob->dx2, prob->x);
  c_math_add(prob->p, attr->step, prob->dy2, prob->y);
  c_math_add(prob->m, attr->step, prob->dz2, prob->z);
  c_math_add(prob->m, attr->step, prob->dsaff, prob->s);

  attr->kap += attr->step * dkap;
  attr->tau += attr->step * dtau;

#ifdef CSOCP_WITH_DEBUG
  c_fprint(prob->dbgfile, "[%s:%d]: Finish corrector step (%.4fs)\n", __FILE__,
           __LINE__, c_timer_elapsed() - timebeg);
  c_fprint(prob->dbgfile, "\n");
#endif

  return;
}

/* Get memory requirement of CSOCP solver */
c_int c_solver_getmem(c_int n,    /* Number of variables */
                      c_int m,    /* Number of inequalities */
                      c_int p,    /* Number of equalities */
                      c_int nnzA, /* Number of nonzeros of A */
                      c_int nnzG  /* Number of nonzeros of G */
) {
  c_int probsize = 0;

  /* Memory of problem structure */
  probsize += sizeof(csocp_prob);

  /* Memory of problem data */
#ifdef CSOCP_WITH_COPYPROB
  probsize += c_spmat_getmem(n, nnzA); /* A */
  probsize += c_spmat_getmem(n, nnzG); /* G */
  probsize += n * sizeof(c_real);      /* c */
  probsize += m * sizeof(c_real);      /* b */
  probsize += p * sizeof(c_real);      /* h */
#else
  probsize += sizeof(c_spmat *); /* A */
  probsize += sizeof(c_spmat *); /* G */
  probsize += sizeof(c_real *);  /* c */
  probsize += sizeof(c_real *);  /* b */
  probsize += sizeof(c_real *);  /* h */
#endif

  probsize += sizeof(c_cone *);   /* cone */
  probsize += sizeof(c_kkt *);    /* kkt */
  probsize += sizeof(c_param *);  /* param */
  probsize += sizeof(c_buffer *); /* buffer */

  probsize += n * sizeof(c_real); /* facX */
  probsize += m * sizeof(c_real); /* facA */
  probsize += p * sizeof(c_real); /* facG */

  probsize += n * sizeof(c_real); /* rx */
  probsize += m * sizeof(c_real); /* ry */
  probsize += p * sizeof(c_real); /* rz */

  probsize += n * sizeof(c_real); /* dx1 */
  probsize += m * sizeof(c_real); /* dy1 */
  probsize += p * sizeof(c_real); /* dz1 */

  probsize += n * sizeof(c_real); /* dx2 */
  probsize += m * sizeof(c_real); /* dy2 */
  probsize += p * sizeof(c_real); /* dz2 */

  probsize += m * sizeof(c_real); /* dsaff */
  probsize += m * sizeof(c_real); /* dzaff */
  probsize += m * sizeof(c_real); /* dsaff_w */
  probsize += m * sizeof(c_real); /* dzaff_w */

  probsize += m * sizeof(c_real); /* lambda */

  probsize += n * sizeof(c_real); /* x */
  probsize += p * sizeof(c_real); /* y */
  probsize += m * sizeof(c_real); /* z */
  probsize += m * sizeof(c_real); /* s */
  probsize += sizeof(c_attr *);   /* attr */

  probsize += n * sizeof(c_real); /* best_x */
  probsize += p * sizeof(c_real); /* best_y */
  probsize += m * sizeof(c_real); /* best_z */
  probsize += m * sizeof(c_real); /* best_s */
  probsize += sizeof(c_attr *);   /* best_attr */

  probsize += sizeof(c_real); /* resx0 */
  probsize += sizeof(c_real); /* resy0 */
  probsize += sizeof(c_real); /* resz0 */

  return probsize;
}
