/**
 * This file implements routines for dense vector in CSOCP solver
 *
 * This module serves as part of dense linear algebra module.
 */

#include "csocp_config.h"

#include "tool_math.h"

#include <math.h>
#include <stdio.h>

/* Dense vector addition */
void c_math_add(c_int nVec,       /* Dimension of vector */
                c_real dFactor,   /* Scaling factor or offset value */
                c_real *srcVec,   /* Input dense vector */
                c_real *srcDstVec /* Output + Input dense vector */
) {
  c_int iVec = 0;

  /* If dFactor is 0.0, no computation will be performed */
  if (dFactor == 0.0) {
    return;
  }

  /* If srcVec == NULL:
   *   srcDstVec += dFactor
   * Else:
   *   srcDstVec += dFactor * srcVec
   */
  if (srcVec == NULL) {
    for (iVec = 0; iVec < nVec; ++iVec) {
      srcDstVec[iVec] += dFactor;
    }
  } else {
    if (dFactor == 1.0) {
      for (iVec = 0; iVec < nVec; ++iVec) {
        srcDstVec[iVec] += srcVec[iVec];
      }
    } else {
      for (iVec = 0; iVec < nVec; ++iVec) {
        srcDstVec[iVec] += dFactor * srcVec[iVec];
      }
    }
  }

  return;
}

/* Dense vector subtraction */
void c_math_sub(c_int nVec,       /* Dimension of vector */
                c_real dFactor,   /* Scaling factor or offset value */
                c_real *srcVec,   /* Input dense vector */
                c_real *srcDstVec /* Output + Input dense vector */
) {
  c_int iVec = 0;

  /* If dFactor is 0.0, no computation will be performed */
  if (dFactor == 0.0) {
    return;
  }

  /* If srcVec == NULL:
   *   srcDstVec -= dFactor
   * Else:
   *   srcDstVec -= dFactor * srcVec
   */
  if (srcVec == NULL) {
    for (iVec = 0; iVec < nVec; ++iVec) {
      srcDstVec[iVec] -= dFactor;
    }
  } else {
    if (dFactor == 1.0) {
      for (iVec = 0; iVec < nVec; ++iVec) {
        srcDstVec[iVec] -= srcVec[iVec];
      }
    } else {
      for (iVec = 0; iVec < nVec; ++iVec) {
        srcDstVec[iVec] -= dFactor * srcVec[iVec];
      }
    }
  }

  return;
}

/* Dense vector multiplication */
void c_math_mul(c_int nVec,       /* Dimension of vector */
                c_real *srcVec,   /* Input dense vector */
                c_real *srcDstVec /* Output + Input dense vector */
) {
  c_int iVec = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    srcDstVec[iVec] *= srcVec[iVec];
  }
  return;
}

/* Dense vector division */
void c_math_div(c_int nVec,       /* Dimension of vector */
                c_real *srcVec,   /* Input dense vector */
                c_real *srcDstVec /* Output + Input dense vector */
) {
  c_int iVec = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
#ifdef CSOCP_WITH_C6678
    srcDstVec[iVec] = c_math_divd(srcDstVec[iVec], srcVec[iVec]);
#else
    srcDstVec[iVec] /= srcVec[iVec];
#endif
  }
  return;
}

/* Scale dense vector by multiplication */
void c_math_scalemul(c_int nVec,       /* Dimension of vector */
                     c_real dFactor,   /* Scaling factor */
                     c_real *srcDstVec /* Output + Input dense vector */
) {
  c_int iVec = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    srcDstVec[iVec] *= dFactor;
  }
  return;
}

/* Scale dense vector by division */
void c_math_scalediv(c_int nVec,       /* Dimension of vector */
                     c_real dFactor,   /* Scaling factor */
                     c_real *srcDstVec /* Output + Input dense vector */
) {
  c_int iVec = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
#ifdef CSOCP_WITH_C6678
    srcDstVec[iVec] = c_math_divd(srcDstVec[iVec], dFactor);
#else
    srcDstVec[iVec] /= dFactor;
#endif
  }
  return;
}

/* Scale dense vector by index */
void c_math_scalewithidx(c_int nVec,       /* Dimension of vector */
                         c_real dFactor,   /* Scaling factor */
                         c_int *srcIdx,    /* Input index */
                         c_real *srcDstVec /* Output + Input dense vector */
) {
  c_int iVec = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    srcDstVec[srcIdx[iVec]] *= dFactor;
  }
  return;
}

/* Copy double dense vector */
void c_math_copydbl(c_int nVec,     /* Dimension of vector */
                    c_real *srcVec, /* Input dense vector */
                    c_real *dstVec, /* Output dense vector */
                    c_int ifNegate  /* Whether to negate values */
) {
  c_int iVec = 0;
  if (ifNegate) {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[iVec] = -srcVec[iVec];
    }
  } else {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[iVec] = srcVec[iVec];
    }
  }
  return;
}

/* Copy integer dense vector */
void c_math_copyint(c_int nVec,    /* Dimension of vector */
                    c_int *srcVec, /* Input dense vector */
                    c_int *dstVec, /* Output dense vector */
                    c_int ifNegate /* Whether to negate values */
) {
  c_int iVec = 0;
  if (ifNegate) {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[iVec] = -srcVec[iVec];
    }
  } else {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[iVec] = srcVec[iVec];
    }
  }
  return;
}

/* Copy dense double vector by index */
void c_math_copydblbyidx(c_int nVec,     /* Dimension of vector */
                         c_int *srcIdx,  /* Input index */
                         c_real *srcVec, /* Input dense vector */
                         c_real *dstVec, /* Output dense vector */
                         c_int ifNegate  /* Whether to negate values */
) {
  c_int iVec = 0;
  if (ifNegate) {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[srcIdx[iVec]] = -srcVec[iVec];
    }
  } else {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[srcIdx[iVec]] = srcVec[iVec];
    }
  }
  return;
}

/* Copy dense integer vector by index */
void c_math_copyintbyidx(c_int nVec,    /* Dimension of vector */
                         c_int *srcIdx, /* Input index */
                         c_int *srcVec, /* Input dense vector */
                         c_int *dstVec, /* Output dense vector */
                         c_int ifNegate /* Whether to negate values */
) {
  c_int iVec = 0;
  if (ifNegate) {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[srcIdx[iVec]] = -srcVec[iVec];
    }
  } else {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[srcIdx[iVec]] = srcVec[iVec];
    }
  }
  return;
}

/* Copy two dense vector by operator and index */
void c_math_copywithopidx(c_int nVec,     /* Dimension of vector */
                          c_int *srcIdx,  /* Input index */
                          c_real *srcX,   /* Input dense vector */
                          c_real *srcY,   /* Input dense vector */
                          c_real *dstVec, /* Output dense vector */
                          c_int iSign     /* Operator for input vectors */
) {
  c_int iVec = 0;
  if (iSign > 0) {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[srcIdx[iVec]] = srcX[iVec] + srcY[iVec];
    }
  } else {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[srcIdx[iVec]] = srcX[iVec] - srcY[iVec];
    }
  }
  return;
}

/* Copy dense double vector with inverse index */
void c_math_copydblbyinvidx(c_int nVec,       /* Dimension of vector */
                            c_int *srcInvIdx, /* Input inverse index */
                            c_real *srcVec,   /* Input dense vector */
                            c_real *dstVec    /* Output dense vector */
) {
  c_int iVec = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    dstVec[iVec] = srcVec[srcInvIdx[iVec]];
  }
  return;
}

/* Fill vector with given double value */
void c_math_filldbl(c_int nVec,    /* Dimension of vector */
                    c_real srcVal, /* Input double scalar */
                    c_real *dstVec /* Output dense vector */
) {
  c_int iVec = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    dstVec[iVec] = srcVal;
  }
  return;
}

/* Fill vector with given integer value */
void c_math_fillint(c_int nVec,   /* Dimension of vector */
                    c_int srcVal, /* Input integer scalar */
                    c_int *dstVec /* Output dense vector */
) {
  c_int iVec = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    dstVec[iVec] = srcVal;
  }
  return;
}

/* Fill vector by index */
void c_math_fillbyidx(c_int nVec,     /* Dimension of vector */
                      c_real srcVal,  /* Input double scalar */
                      c_int *srcIdx1, /* First level index */
                      c_int *srcIdx2, /* Second level index */
                      c_real *dstVec  /* Output dense vector */
) {
  c_int iVec = 0;
  if (srcIdx1 != NULL && srcIdx2 != NULL) {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[srcIdx1[srcIdx2[iVec]]] = srcVal;
    }
  } else if (srcIdx1 != NULL && srcIdx2 == NULL) {
    for (iVec = 0; iVec < nVec; ++iVec) {
      dstVec[srcIdx1[iVec]] = srcVal;
    }
  }
  return;
}

/* 2-norm of dense vector */
c_real c_math_norm2(c_int nVec,    /* Dimension of vector */
                    c_real *srcVec /* Input dense vector */
) {
  c_int iVec = 0;
  c_real dNormS = 0.0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    dNormS += srcVec[iVec] * srcVec[iVec];
  }
  return c_math_sqrt(dNormS);
}

/* Inf-norm of dense vector */
c_real c_math_norminf(c_int nVec,    /* Dimension of vector */
                      c_real *srcVec /* Input dense vector */
) {
  c_int iVec = 0;
  c_real dInfNorm = 0.0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    dInfNorm = CSOCP_MAX(srcVec[iVec], dInfNorm);
    dInfNorm = CSOCP_MAX(-srcVec[iVec], dInfNorm);
  }
  return dInfNorm;
}

/* Dot product of dense vectors */
c_real c_math_dot(c_int nVec,   /* Dimension of vector */
                  c_real *srcX, /* Input dense vector */
                  c_real *srcY  /* Input dense vector */
) {
  c_real dDot = 0.0;
  c_int iVec = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    dDot += srcX[iVec] * srcY[iVec];
  }
  return dDot;
}

/* Summation of dense vector */
c_real c_math_sum(c_int nVec,    /* Dimension of vector */
                  c_real *srcVec /* Input dense vector */
) {
  c_real dSum = 0;
  c_int iVec = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    dSum += srcVec[iVec];
  }
  return dSum;
}

/* Cumulative summation of dense vector, results stored in both input and output */
void c_math_cumsum(c_int nVec,    /* Dimension of vector */
                   c_int *srcVec, /* Input dense vector */
                   c_int *dstVec  /* Output dense vector */
) {
  c_int iVec = 0;
  c_int dCumSum = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    dstVec[iVec] = dCumSum;
    dCumSum += srcVec[iVec];
    srcVec[iVec] = dstVec[iVec];
  }
  return;
}

/* Inverse index map */
void c_math_invidx(c_int nVec,    /* Dimension of vector */
                   c_int *srcIdx, /* Input index */
                   c_int *dstIdx  /* Output index */
) {
  c_int iVec = 0;
  for (iVec = 0; iVec < nVec; ++iVec) {
    dstIdx[srcIdx[iVec]] = iVec;
  }
  return;
}

/* Division: A / B */
c_real c_math_divd(c_real A, /* Dividend */
                   c_real B  /* Divisor */
) {
#ifdef CSOCP_WITH_C6678
  c_real X0, X1, X2, X3, Y, TWO = 2.0;
  X0 = _rcpdp(B);
  X1 = X0 * (TWO - B * X0);
  X2 = X1 * (TWO - B * X1);
  X3 = X2 * (TWO - B * X2);
  Y = A * X3;
  return Y;
#else
  return A / B;
#endif
}

/* Squared root: sqrt(A) */
c_real c_math_sqrt(c_real A /* Input value */
) {
#ifdef CSOCP_WITH_C6678
  c_real HALF = 0.5, K15 = 1.5, AH, X0, X1, X2, X3, Y;
  AH = A * HALF;
  X0 = _rsqrdp(A);
  X1 = X0 * (K15 - (AH * X0) * X0);
  X2 = X1 * (K15 - (AH * X1) * X1);
  X3 = X2 * (K15 - (AH * X2) * X2);
  Y = A * X3;
  return Y;
#else
  return sqrt(A);
#endif
}

/* Dump double vector to file */
void c_math_dumpdbl(c_int nVec,     /* Length of vector */
                    c_real *srcVec, /* Input dense vector */
                    char *vecFile   /* Dumped vector file */
) {
#ifdef CSOCP_WITH_DUMP
  c_int i = 0;
  FILE *f = NULL;

  if (vecFile == NULL) {
    for (i = 0; i < nVec; ++i) {
      fprintf(stdout, "%6d, %+.17g\n", i, srcVec[i]);
    }
  } else {
    f = fopen(vecFile, "w");
    if (f != NULL) {
      for (i = 0; i < nVec; ++i) {
        fprintf(f, "%6d, %+.17g\n", i, srcVec[i]);
      }
      fclose(f);
    }
  }
#endif
  return;
}

/* Dump integer vector to file */
void c_math_dumpint(c_int nVec,    /* Length of vector */
                    c_int *srcVec, /* Input dense vector */
                    char *vecFile  /* Dumped vector file */
) {
#ifdef CSOCP_WITH_DUMP
  c_int i = 0;
  FILE *f = NULL;

  if (vecFile == NULL) {
    for (i = 0; i < nVec; ++i) {
      fprintf(stdout, "%d, %d\n", (int)i, (int)srcVec[i]);
    }
  } else {
    f = fopen(vecFile, "w");
    if (f != NULL) {
      for (i = 0; i < nVec; ++i) {
        fprintf(f, "%d, %d\n", (int)i, (int)srcVec[i]);
      }
      fclose(f);
    }
  }
#endif
  return;
}
