/**
 * This file is function prototypes for dense vector in CSOCP solver
 *
 * This module serves as part of dense linear algebra module.
 */

#ifndef __TOOL_MATH_H__
#define __TOOL_MATH_H__

#include "csocp_config.h"

#define CSOCP_MIN(X, Y) ((X) < (Y) ? (X) : (Y))
#define CSOCP_MAX(X, Y) ((X) > (Y) ? (X) : (Y))

#ifndef CSOCP_WITH_C6678
#include <math.h>
#define CSOCP_ABS(X) fabs(X)
#else
#define CSOCP_ABS(X) ((X) < 0 ? -(X) : (X))
#endif

/* Dense vector addition */
void c_math_add(c_int nVec,       /* Dimension of vector */
                c_real dFactor,   /* Scaling factor or offset value */
                c_real *srcVec,   /* Input dense vector */
                c_real *srcDstVec /* Output + Input dense vector */
);

/* Dense vector subtraction */
void c_math_sub(c_int nVec,       /* Dimension of vector */
                c_real dFactor,   /* Scaling factor or offset value */
                c_real *srcVec,   /* Input dense vector */
                c_real *srcDstVec /* Output + Input dense vector */
);

/* Dense vector multiplication */
void c_math_mul(c_int nVec,       /* Dimension of vector */
                c_real *srcVec,   /* Input dense vector */
                c_real *srcDstVec /* Output + Input dense vector */
);

/* Dense vector division */
void c_math_div(c_int nVec,       /* Dimension of vector */
                c_real *srcVec,   /* Input dense vector */
                c_real *srcDstVec /* Output + Input dense vector */
);

/* Scale dense vector by multiplication */
void c_math_scalemul(c_int nVec,       /* Dimension of vector */
                     c_real dFactor,   /* Scaling factor */
                     c_real *srcDstVec /* Output + Input dense vector */
);

/* Scale dense vector by division */
void c_math_scalediv(c_int nVec,       /* Dimension of vector */
                     c_real dFactor,   /* Scaling factor */
                     c_real *srcDstVec /* Output + Input dense vector */
);

/* Scale dense vector by index */
void c_math_scalewithidx(c_int nVec,       /* Dimension of vector */
                         c_real dFactor,   /* Scaling factor */
                         c_int *srcIdx,    /* Input index */
                         c_real *srcDstVec /* Output + Input dense vector */
);

/* Copy double dense vector */
void c_math_copydbl(c_int nVec,     /* Dimension of vector */
                    c_real *srcVec, /* Input dense vector */
                    c_real *dstVec, /* Output dense vector */
                    c_int ifNegate  /* Whether to negate values */
);

/* Copy integer dense vector */
void c_math_copyint(c_int nVec,    /* Dimension of vector */
                    c_int *srcVec, /* Input dense vector */
                    c_int *dstVec, /* Output dense vector */
                    c_int ifNegate /* Whether to negate values */
);

/* Copy dense double vector by index */
void c_math_copydblbyidx(c_int nVec,     /* Dimension of vector */
                         c_int *srcIdx,  /* Input index */
                         c_real *srcVec, /* Input dense vector */
                         c_real *dstVec, /* Output dense vector */
                         c_int ifNegate  /* Whether to negate values */
);

/* Copy dense integer vector by index */
void c_math_copyintbyidx(c_int nVec,    /* Dimension of vector */
                         c_int *srcIdx, /* Input index */
                         c_int *srcVec, /* Input dense vector */
                         c_int *dstVec, /* Output dense vector */
                         c_int ifNegate /* Whether to negate values */
);

/* Copy two dense vector by operator and index */
void c_math_copywithopidx(c_int nVec,     /* Dimension of vector */
                          c_int *srcIdx,  /* Input index */
                          c_real *srcX,   /* Input dense vector */
                          c_real *srcY,   /* Input dense vector */
                          c_real *dstVec, /* Output dense vector */
                          c_int iSign     /* Operator for input vectors */
);

/* Copy dense double vector with inverse index */
void c_math_copydblbyinvidx(c_int nVec,       /* Dimension of vector */
                            c_int *srcInvIdx, /* Input inverse index */
                            c_real *srcVec,   /* Input dense vector */
                            c_real *dstVec    /* Output dense vector */
);

/* Fill vector with given double value */
void c_math_filldbl(c_int nVec,    /* Dimension of vector */
                    c_real srcVal, /* Input double scalar */
                    c_real *dstVec /* Output dense vector */
);

/* Fill vector with given integer value */
void c_math_fillint(c_int nVec,   /* Dimension of vector */
                    c_int srcVal, /* Input integer scalar */
                    c_int *dstVec /* Output dense vector */
);

/* Fill vector by index */
void c_math_fillbyidx(c_int nVec,     /* Dimension of vector */
                      c_real srcVal,  /* Input double scalar */
                      c_int *srcIdx1, /* First level index */
                      c_int *srcIdx2, /* Second level index */
                      c_real *dstVec  /* Output dense vector */
);

/* 2-norm of dense vector */
c_real c_math_norm2(c_int nVec,    /* Dimension of vector */
                    c_real *srcVec /* Input dense vector */
);

/* Inf-norm of dense vector */
c_real c_math_norminf(c_int nVec,    /* Dimension of vector */
                      c_real *srcVec /* Input dense vector */
);

/* Dot product of dense vectors */
c_real c_math_dot(c_int nVec,   /* Dimension of vector */
                  c_real *srcX, /* Input dense vector */
                  c_real *srcY  /* Input dense vector */
);

/* Summation of dense vector */
c_real c_math_sum(c_int nVec,    /* Dimension of vector */
                  c_real *srcVec /* Input dense vector */
);

/* Cumulative summation of dense vector, results stored in both input and output */
void c_math_cumsum(c_int nVec,    /* Dimension of vector */
                   c_int *srcVec, /* Input dense vector */
                   c_int *dstVec  /* Output dense vector */
);

/* Inverse index map */
void c_math_invidx(c_int nVec,    /* Dimension of vector */
                   c_int *srcIdx, /* Input index */
                   c_int *dstIdx  /* Output index */
);

/**
 * Special division and squared root
 */

/* Division: A / B */
c_real c_math_divd(c_real A, /* Dividend */
                   c_real B  /* Divisor */
);

/* Squared root: sqrt(A) */
c_real c_math_sqrt(c_real A /* Input value */
);

#ifdef CSOCP_WITH_C6678
#define CSOCP_DIV(X, Y) ((Y) < 1e-13 ? ((X)*1e13) : c_math_divd((X), (Y)))
#else
#define CSOCP_DIV(X, Y) ((Y) < 1e-13 ? ((X) / 1e-13) : (X) / (Y))
#endif

/**
 * Dump methods
 */

/* Dump double vector to file */
void c_math_dumpdbl(c_int nVec,     /* Length of vector */
                    c_real *srcVec, /* Input dense vector */
                    char *vecFile   /* Dumped vector file */
);

/* Dump integer vector to file */
void c_math_dumpint(c_int nVec,    /* Length of vector */
                    c_int *srcVec, /* Input dense vector */
                    char *vecFile  /* Dumped vector file */
);

#endif
