/**
 * This file implements routines for sparse matrix in CSOCP solver
 *
 * This module serves as part of sparse linear algebra module.
 */

#include "csocp_config.h"

#include "tool_buffer.h"
#include "tool_math.h"
#include "tool_memory.h"
#include "tool_spmat.h"

#include "struct_spmat.h"

#include <stdio.h>
#include <stdlib.h>

/**
 * Creating and deleting methods
 */

/* Create a new sparse matrix */
c_spmat *c_spmat_create(c_int nRow, /* Number of rows */
                        c_int nCol, /* Number of columns */
                        c_int nElem /* Number of elements */
) {
  c_int *colBeg = NULL, *colIdx = NULL;
  c_real *colElem = NULL;

  /* Allocate memory for sparse matrix data */
  CSOCP_INIT(colBeg, nCol + 1);
  CSOCP_INIT(colIdx, nElem);
  CSOCP_INIT(colElem, nElem);

  /* Generate a new sparse matrix */
  return c_spmat_generate(nRow, nCol, nElem, colBeg, colIdx, colElem);
}

/* Generate a sparse matrix from existing arrays */
c_spmat *c_spmat_generate(c_int nRow,     /* Number of rows */
                          c_int nCol,     /* Number of columns */
                          c_int nElem,    /* Number of elements */
                          c_int *colBeg,  /* Begin pointer of column-wise matrix */
                          c_int *colIdx,  /* Index pointer of column-wise matrix */
                          c_real *colElem /* Element pointer of column-wise matrix */
) {
  c_spmat *newMat = NULL;

  /* Allocate memory for sparse matrix */
  CSOCP_INIT(newMat, 1);

  /* Assign dimension of sparse matrix */
  newMat->nRow = nRow;
  newMat->nCol = nCol;
  newMat->nElem = nElem;

  /* Assign sparse matrix data */
  newMat->colBeg = colBeg;
  newMat->colIdx = colIdx;
  newMat->colElem = colElem;

  return newMat;
}

/* Load data to sparse matrix */
void c_spmat_loaddata(c_spmat *dstMat, /* Matrix to load data */
                      c_int *colBeg,   /* Begin pointer of column-wise matrix */
                      c_int *colIdx,   /* Index pointer of column-wise matrix */
                      c_real *colElem  /* Element pointer of column-wise matrix */
) {
  c_int nCol = dstMat->nCol;
  c_int nElem = dstMat->nElem;

  /* Copy sparse matrix data */
  c_math_copyint(nCol + 1, colBeg, dstMat->colBeg, 0);
  c_math_copyint(nElem, colIdx, dstMat->colIdx, 0);
  c_math_copydbl(nElem, colElem, dstMat->colElem, 0);

  return;
}

/* Free a sparse matrix */
void c_spmat_free(c_spmat **p_Mat) {
  if (p_Mat != NULL && *p_Mat != NULL) {
    /* Free sparse matrix data */
    CSOCP_FREE((*p_Mat)->colBeg);
    CSOCP_FREE((*p_Mat)->colIdx);
    CSOCP_FREE((*p_Mat)->colElem);

    CSOCP_FREE(*p_Mat);
  }
  return;
}

/**
 * Advanced operation methods
 */

/* Compute inf-norm for rows/columns of sparse matrix */
void c_spmat_infnorm(c_spmat *srcMat, /* Input sparse matrix */
                     c_real *dstVal,  /* Output values */
                     c_int ifCol      /* Whether to compute for rows/columns */
) {
  c_int i = 0, j = 0, k = 0;

  /* Note: 'dstVal' must be initialized at first call */
  if (ifCol) {
    /* Loop through matrix to compute inf-norm of columns */
    for (i = 0; i < srcMat->nCol; ++i) {
      for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
        dstVal[i] = CSOCP_MAX(CSOCP_ABS(srcMat->colElem[j]), dstVal[i]);
      }
    }
  } else {
    /* Loop through matrix to compute inf-norm of rows */
    for (i = 0; i < srcMat->nCol; ++i) {
      for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
        k = srcMat->colIdx[j];
        dstVal[k] = CSOCP_MAX(CSOCP_ABS(srcMat->colElem[j]), dstVal[k]);
      }
    }
  }

  return;
}

/* Compute 1-norm for rows/columns of sparse matrix */
void c_spmat_onenorm(c_spmat *srcMat, /* Input sparse matrix */
                     c_real *dstVal,  /* Output values */
                     c_int ifCol      /* Whether to compute for rows/columns */
) {
  c_int i = 0, j = 0;

  /* Note: 'dstVal' must be initialized at first call */
  if (ifCol) {
    /* Loop through matrix to compute 1-norm of columns */
    for (i = 0; i < srcMat->nCol; ++i) {
      for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
        dstVal[i] += CSOCP_ABS(srcMat->colElem[j]);
      }
    }
  } else {
    /* Loop through matrix to compute 1-norm of rows */
    for (i = 0; i < srcMat->nCol; ++i) {
      for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
        dstVal[srcMat->colIdx[j]] += CSOCP_ABS(srcMat->colElem[j]);
      }
    }
  }

  return;
}

/* Scale rows/columns of sparse matrix by dividing the factors */
void c_spmat_setscaling(c_spmat *srcMat,   /* Input sparse matrix */
                        c_real *srcFactor, /* Input scaling factors */
                        c_int ifCol        /* Whether to compute for rows/columns */
) {
  c_int i = 0, j = 0;

  if (ifCol) {
    /* Loop through matrix to scale columns */
    for (i = 0; i < srcMat->nCol; ++i) {
      for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
#ifdef CSOCP_WITH_C6678
        srcMat->colElem[j] = c_math_divd(srcMat->colElem[j], srcFactor[i]);
#else
        srcMat->colElem[j] /= srcFactor[i];
#endif
      }
    }
  } else {
    /* Loop through matrix to scale rows */
    for (i = 0; i < srcMat->nCol; ++i) {
      for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
#ifdef CSOCP_WITH_C6678
        srcMat->colElem[j] =
            c_math_divd(srcMat->colElem[j], srcFactor[srcMat->colIdx[j]]);
#else
        srcMat->colElem[j] /= srcFactor[srcMat->colIdx[j]];
#endif
      }
    }
  }

  return;
}

/* Unscale sparse matrix by multiplying the factors */
void c_spmat_unsetscaling(c_spmat *srcMat,   /* Input sparse matrix */
                          c_real *rowFactor, /* Input row scaling factors */
                          c_real *colFactor  /* Input column scaling factors */
) {
  c_int i = 0, j = 0, k = 0;

  /* Loop through matrix to scale columns */
  for (i = 0; i < srcMat->nCol; ++i) {
    for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
      k = srcMat->colIdx[j];
      srcMat->colElem[j] *= (rowFactor[k] * colFactor[i]);
    }
  }
}

/* Transpose a sparse matrix and return transposed sparse matrix */
c_spmat *c_spmat_transpose(c_spmat *srcMat, /* Input sparse matrix */
                           c_buffer *buffer /* Input buffer */
) {
  c_int i = 0, j = 0, k = 0;
  c_int *rowCnt = NULL;
  c_spmat *newMat = NULL;

  /* Create transposed sparse matrix */
  newMat = c_spmat_create(srcMat->nCol, srcMat->nRow, srcMat->nElem);
  if (srcMat->nElem == 0) {
    return newMat;
  }

  /* Obtain temporary counter for rows from buffer */
  c_buffer_request_ibuf(buffer, rowCnt, srcMat->nRow);
  c_math_fillint(srcMat->nRow, 0, rowCnt);

  /* Count number of elements for each row */
  for (j = 0; j < srcMat->nElem; ++j) {
    rowCnt[srcMat->colIdx[j]]++;
  }

  /* Compute 'colBeg' of transposed sparse matrix */
  c_math_cumsum(srcMat->nRow, rowCnt, newMat->colBeg);
  newMat->colBeg[newMat->nCol] = newMat->nElem;

  /* Copy sparse matrix data to transposed sparse matrix */
  for (i = 0; i < srcMat->nCol; ++i) {
    for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
      k = rowCnt[srcMat->colIdx[j]]++;
      newMat->colIdx[k] = i;
      newMat->colElem[k] = srcMat->colElem[j];
    }
  }

  /* Release buffer */
  c_buffer_release_ibuf(buffer, srcMat->nRow);

  return newMat;
}

/**
 * Permutes a symmetric matrix with only the upper triangular part stored
 * So that: dstMat[srcDstMap[k]] = srcMat[k]
 */
void c_spmat_permute(c_spmat *srcMat,  /* Input sparse matrix */
                     c_int *srcInvP,   /* Input inverse permutation */
                     c_spmat *dstMat,  /* Output upper triangular sparse matrix */
                     c_int *srcDstMap, /* Index map between input and output index */
                     c_buffer *buffer  /* Input buffer */
) {
  c_int i = 0, ii = 0, j = 0, jj = 0, k = 0, l = 0;
  c_int *colCnt = NULL;

  /* Obtain temporary counter for columns from buffer */
  c_buffer_request_ibuf(buffer, colCnt, srcMat->nCol);
  c_math_fillint(srcMat->nCol, 0, colCnt);

  /* Counter number of elements for each column */
  for (j = 0; j < srcMat->nCol; ++j) {
    jj = srcInvP[j];
    for (k = srcMat->colBeg[j]; k < srcMat->colBeg[j + 1]; ++k) {
      i = srcMat->colIdx[k];
      /* Count upper triangular submatrix only */
      if (i > j) {
        continue;
      }
      ii = srcInvP[i];
      colCnt[CSOCP_MAX(ii, jj)]++;
    }
  }

  /* Compute 'colBeg' of permuted sparse matrix */
  c_math_cumsum(srcMat->nCol, colCnt, dstMat->colBeg);
  dstMat->colBeg[dstMat->nCol] = dstMat->nElem;

  /* Copy sparse matrix data */
  for (j = 0; j < srcMat->nCol; ++j) {
    jj = srcInvP[j];
    for (k = srcMat->colBeg[j]; k < srcMat->colBeg[j + 1]; ++k) {
      i = srcMat->colIdx[k];
      /* Copy upper triangular submatrix only */
      if (i > j) {
        continue;
      }
      ii = srcInvP[i];

      l = colCnt[CSOCP_MAX(ii, jj)]++;
      dstMat->colIdx[l] = CSOCP_MIN(ii, jj);
      dstMat->colElem[l] = srcMat->colElem[k];
      if (srcDstMap != NULL) {
        srcDstMap[k] = l;
      }
    }
  }

  /* Release buffer */
  c_buffer_release_ibuf(buffer, srcMat->nCol);

  return;
}

/**
 * Sparse symmetric matrix-vector multiplication
 *   1. srcDstY  =  srcMat*srcX, nSign > 0 && ifNewVec == 1
 *   2. srcDstY +=  srcMat*srcX, nSign > 0 && ifNewVec == 0
 *   3. srcDstY  = -srcMat*srcX, nSign < 0 && ifNewVec == 1
 *   4, srcDstY -=  srcMat*srcX, nSign < 0 && ifNewVec == 0
 */
void c_spmat_matvec(c_spmat *srcMat, /* Input sparse matrix */
                    c_real *srcX,    /* Input dense vector */
                    c_real *srcDstY, /* Output + Input dense vector */
                    c_int nSign,     /* Whether to do addition or subtraction */
                    c_int ifNewVec   /* Whether to use vector Y for output only */
) {
  c_int i = 0, j = 0;

  /* Skip calculation if sparse matrix is empty */
  if (srcMat->nElem == 0) {
    return;
  }

  /* Initialize vector Y to zeros since it is used as output only */
  if (ifNewVec > 0) {
    c_math_filldbl(srcMat->nRow, 0.0, srcDstY);
  }

  if (nSign > 0) {
    /* Do summation */
    for (i = 0; i < srcMat->nCol; ++i) {
      for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
        srcDstY[srcMat->colIdx[j]] += srcMat->colElem[j] * srcX[i];
      }
    }
  } else {
    /* Do subtraction */
    for (i = 0; i < srcMat->nCol; ++i) {
      for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
        srcDstY[srcMat->colIdx[j]] -= srcMat->colElem[j] * srcX[i];
      }
    }
  }

  return;
}

/**
 * Sparse symmetric matrix-transpose-vector multiplication:
 *   1. srcDstY  = -srcMat'*srcX, ifNewVec > 0
 *   2. srcDstY -=  srcMat'*srcX, ifNewVec = 0
 */
void c_spmat_mattvec(c_spmat *srcMat, /* Input sparse matrix */
                     c_real *srcX,    /* Input dense vector */
                     c_real *srcDstY, /* Output + Input dense vector */
                     c_int ifNewVec,  /* Whether to use vector Y for output only */
                     c_int ifSkipDiag /* Whether to skip diagonal elements */
) {
  c_int i = 0, j = 0;

  /* Skip calculation if sparse matrix is empty */
  if (srcMat->nElem == 0) {
    return;
  }

  /* Initialize vector Y to zeros since it is used as output only */
  if (ifNewVec > 0) {
    c_math_filldbl(srcMat->nCol, 0.0, srcDstY);
  }

  if (ifSkipDiag) {
    /* Do subtraction with diagonal elements ignored */
    for (i = 0; i < srcMat->nCol; ++i) {
      for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
        if (srcMat->colIdx[j] != i) {
          srcDstY[i] -= srcMat->colElem[j] * srcX[srcMat->colIdx[j]];
        }
      }
    }
  } else {
    /* Do subtraction */
    for (i = 0; i < srcMat->nCol; ++i) {
      for (j = srcMat->colBeg[i]; j < srcMat->colBeg[i + 1]; ++j) {
        srcDstY[i] -= srcMat->colElem[j] * srcX[srcMat->colIdx[j]];
      }
    }
  }

  return;
}

/**
 * Auxiliary methods
 */

/* Dump sparse matrix to file for debug purpose */
void c_spmat_dump(c_spmat *srcMat, /* Input sparse matrix */
                  char *matFile    /* Dumped matrix file */
) {
#ifdef CSOCP_WITH_DUMP
  c_int j = 0, k = 0;

  FILE *f = fopen(matFile, "w");
  if (f != NULL) {
    for (j = 0; j < srcMat->nCol; j++) {
      for (k = srcMat->colBeg[j]; k < srcMat->colBeg[j + 1]; k++) {
        fprintf(f, "%6d %6d %+.17g\n", srcMat->colIdx[k], j, srcMat->colElem[k]);
      }
    }
    fclose(f);
  }
#endif
  return;
}

/* Get memory requirement of sparse matrix */
c_int c_spmat_getmem(c_int nCol, /* Number of columns */
                     c_int nElem /* Number of elements */
) {
  c_int matsize = 0;

  /* Memory of sparse matrix structure */
  matsize += sizeof(c_spmat);

  /* Memory of sparse matrix data */
  matsize += (nCol + 1) * sizeof(c_int); /* colBeg */
  matsize += nElem * sizeof(c_int);      /* colIdx */
  matsize += nElem * sizeof(c_real);     /* colElem */

  return matsize;
}
