/**
 * This is function prototypes for sparse matrix in CSOCP solver
 *
 * This module serves as part of sparse linear algebra module.
 */

#ifndef __TOOL_SPMAT_H__
#define __TOOL_SPMAT_H__

#include "csocp_config.h"

/**
 * Creating and deleting methods
 */

/* Create a new sparse matrix */
c_spmat *c_spmat_create(c_int nRow, /* Number of rows */
                        c_int nCol, /* Number of columns */
                        c_int nElem /* Number of elements */
);

/* Generate a sparse matrix from existing arrays */
c_spmat *c_spmat_generate(c_int nRow,     /* Number of rows */
                          c_int nCol,     /* Number of columns */
                          c_int nElem,    /* Number of elements */
                          c_int *colBeg,  /* Begin pointer of column-wise matrix */
                          c_int *colIdx,  /* Index pointer of column-wise matrix */
                          c_real *colElem /* Element pointer of column-wise matrix */
);

/* Load data to sparse matrix */
void c_spmat_loaddata(c_spmat *dstMat, /* Matrix to load data */
                      c_int *colBeg,   /* Begin pointer of column-wise matrix */
                      c_int *colIdx,   /* Index pointer of column-wise matrix */
                      c_real *colElem  /* Element pointer of column-wise matrix */
);

/* Free a sparse matrix */
void c_spmat_free(c_spmat **p_Mat);

/**
 * Advanced operation methods
 */

/* Compute inf-norm for rows/columns of sparse matrix */
void c_spmat_infnorm(c_spmat *srcMat, /* Input sparse matrix */
                     c_real *dstVal,  /* Output values */
                     c_int ifCol      /* Whether to compute for rows/columns */
);

/* Compute 1-norm for rows/columns of sparse matrix */
void c_spmat_onenorm(c_spmat *srcMat, /* Input sparse matrix */
                     c_real *dstVal,  /* Output values */
                     c_int ifCol      /* Whether to compute for rows/columns */
);

/* Scale rows/columns of sparse matrix by dividing the factors */
void c_spmat_setscaling(c_spmat *srcMat,   /* Input sparse matrix */
                        c_real *srcFactor, /* Input scaling factors */
                        c_int ifCol        /* Whether to compute for rows/columns */
);

/* Unscale sparse matrix by multiplying the factors */
void c_spmat_unsetscaling(c_spmat *srcMat,   /* Input sparse matrix */
                          c_real *rowFactor, /* Input row scaling factors */
                          c_real *colFactor  /* Input column scaling factors */
);

/* Transpose a sparse matrix and return transposed sparse matrix */
c_spmat *c_spmat_transpose(c_spmat *srcMat, /* Input sparse matrix */
                           c_buffer *buffer /* Input buffer */
);

/**
 * Permutes a symmetric matrix with only the upper triangular part stored
 * So that: dstMat[srcDstMap[k]] = srcMat[k]
 */
void c_spmat_permute(c_spmat *srcMat,  /* Input sparse matrix */
                     c_int *srcInvP,   /* Input inverse permutation */
                     c_spmat *dstMat,  /* Output upper triangular sparse matrix */
                     c_int *srcDstMap, /* Index map between input and output index */
                     c_buffer *buffer  /* Input buffer */
);

/**
 * Sparse matrix-vector multiplication
 *   1. srcDstY  =  srcMat*srcX, nSign > 0 && ifNewVec == 1
 *   2. srcDstY +=  srcMat*srcX, nSign > 0 && ifNewVec == 0
 *   3. srcDstY  = -srcMat*srcX, nSign < 0 && ifNewVec == 1
 *   4, srcDstY -=  srcMat*srcX, nSign < 0 && ifNewVec == 0
 */
void c_spmat_matvec(c_spmat *srcMat, /* Input sparse matrix */
                    c_real *srcX,    /* Input dense vector */
                    c_real *srcDstY, /* Output + Input dense vector */
                    c_int nSign,     /* Whether to do addition or subtraction */
                    c_int ifNewVec   /* Whether to use vector Y for output only */
);

/**
 * Sparse symmetric matrix-transpose-vector multiplication:
 *   1. srcDstY  = -srcMat'*srcX, ifNewVec > 0
 *   2. srcDstY -=  srcMat'*srcX, ifNewVec = 0
 */
void c_spmat_mattvec(c_spmat *srcMat, /* Input sparse matrix */
                     c_real *srcX,    /* Input dense vector */
                     c_real *srcDstY, /* Output + Input dense vector */
                     c_int ifNewVec,  /* Whether to use vector Y for output only */
                     c_int ifSkipDiag /* Whether to skip diagonal elements */
);

/**
 * Auxiliary methods
 */

/* Dump sparse matrix to file for debug purpose */
void c_spmat_dump(c_spmat *srcMat, /* Input sparse matrix */
                  char *matFile    /* Dumped matrix file */
);

/* Get memory requirement of sparse matrix */
c_int c_spmat_getmem(c_int nCol, /* Number of columns */
                     c_int nElem /* Number of elements */
);

#endif
